# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import struct
import uuid
import logging
import decimal
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Dict,
    Any,
    Union,
    cast,
    TYPE_CHECKING,
)

from typing_extensions import Literal


from .message import Message, Header, Properties

if TYPE_CHECKING:
    from .message import MessageDict

_LOGGER = logging.getLogger(__name__)
_HEADER_PREFIX = memoryview(b"AMQP")
_COMPOSITES = {
    35: "received",
    36: "accepted",
    37: "rejected",
    38: "released",
    39: "modified",
}

c_unsigned_char = struct.Struct(">B")
c_signed_char = struct.Struct(">b")
c_unsigned_short = struct.Struct(">H")
c_signed_short = struct.Struct(">h")
c_unsigned_int = struct.Struct(">I")
c_signed_int = struct.Struct(">i")
c_unsigned_long = struct.Struct(">L")
c_unsigned_long_long = struct.Struct(">Q")
c_signed_long_long = struct.Struct(">q")
c_float = struct.Struct(">f")
c_double = struct.Struct(">d")

DECIMAL128_EXPONENT_MAX = 6144
DECIMAL128_EXPONENT_MIN = -6143
DECIMAL128_MAX_DIGITS = 34
DECIMAL128_BIAS = 6176



def _decode_null(buffer: memoryview) -> Tuple[memoryview, None]:
    return buffer, None


def _decode_true(buffer: memoryview) -> Tuple[memoryview, Literal[True]]:
    return buffer, True


def _decode_false(buffer: memoryview) -> Tuple[memoryview, Literal[False]]:
    return buffer, False


def _decode_zero(buffer: memoryview) -> Tuple[memoryview, Literal[0]]:
    return buffer, 0


def _decode_empty(buffer: memoryview) -> Tuple[memoryview, List[Any]]:
    return buffer, []


def _decode_boolean(buffer: memoryview) -> Tuple[memoryview, bool]:
    return buffer[1:], buffer[:1] == b"\x01"


def _decode_ubyte(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], buffer[0]


def _decode_ushort(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[2:], c_unsigned_short.unpack(buffer[:2])[0]


def _decode_uint_small(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], buffer[0]


def _decode_uint_large(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[4:], c_unsigned_int.unpack(buffer[:4])[0]


def _decode_ulong_small(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], buffer[0]


def _decode_ulong_large(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[8:], c_unsigned_long_long.unpack(buffer[:8])[0]


def _decode_byte(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], c_signed_char.unpack(buffer[:1])[0]


def _decode_short(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[2:], c_signed_short.unpack(buffer[:2])[0]


def _decode_int_small(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], c_signed_char.unpack(buffer[:1])[0]


def _decode_int_large(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[4:], c_signed_int.unpack(buffer[:4])[0]


def _decode_long_small(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[1:], c_signed_char.unpack(buffer[:1])[0]


def _decode_long_large(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[8:], c_signed_long_long.unpack(buffer[:8])[0]


def _decode_float(buffer: memoryview) -> Tuple[memoryview, float]:
    return buffer[4:], c_float.unpack(buffer[:4])[0]


def _decode_double(buffer: memoryview) -> Tuple[memoryview, float]:
    return buffer[8:], c_double.unpack(buffer[:8])[0]


def _decode_timestamp(buffer: memoryview) -> Tuple[memoryview, int]:
    return buffer[8:], c_signed_long_long.unpack(buffer[:8])[0]


def _decode_uuid(buffer: memoryview) -> Tuple[memoryview, uuid.UUID]:
    return buffer[16:], uuid.UUID(bytes=buffer[:16].tobytes())


def _decode_binary_small(buffer: memoryview) -> Tuple[memoryview, bytes]:
    length_index = buffer[0] + 1
    return buffer[length_index:], buffer[1:length_index].tobytes()


def _decode_binary_large(buffer: memoryview) -> Tuple[memoryview, bytes]:
    length_index = c_unsigned_long.unpack(buffer[:4])[0] + 4
    return buffer[length_index:], buffer[4:length_index].tobytes()

def _decode_decimal128(buffer: memoryview) -> Tuple[memoryview, decimal.Decimal]:
    """
    Decode a Decimal128 value from the buffer.

    The Decimal128 encoding is a 16-byte encoding that represents a 34-digit decimal number.
    The encoding uses the IEEE 754-2008 decimal128 format.
    See: https://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-types-v1.0-os.html#type-decimal128

    :param buffer: The buffer containing the Decimal128 encoded value.
    :type buffer: memoryview
    :return: A tuple containing the remaining buffer and the decoded decimal.Decimal value.
    :rtype: Tuple[memoryview, decimal.Decimal]
    """

    # per the spec decimal128 has a width of 16 bytes
    dec_value = bytearray(buffer[:16])

    # Determine the sign of the decimal number
    sign = 1 if dec_value[0] & 0x80 else 0
    biased_exponent = 0

    # Check if the exponent is not a special value
    if dec_value[0] & 0x60 != 0x60:
        # Extract the biased exponent from the first two bytes
        biased_exponent = ((dec_value[0] & 0x7F) << 9) | ((dec_value[1] & 0xFE) >> 1)
        dec_value[0] = 0
        dec_value[1] &= 0x01
    elif dec_value[0] & 0x78 != 0:
        # Handle special values (NaN and Infinity)
        if (dec_value[0] & 0x78) == 0x78:
            return buffer[16:], decimal.Decimal('NaN')
        return buffer[16:], decimal.Decimal('-Infinity')
    else:
        # If the exponent is zero, return zero
        return buffer[16:], decimal.Decimal(0)

    # Calculate the actual exponent by subtracting the bias
    exponent = biased_exponent - DECIMAL128_BIAS

    # Extract the significant digits from the remaining 14 bytes
    hi = c_unsigned_int.unpack(dec_value[4:8])[0]
    middle = c_unsigned_int.unpack(dec_value[8:12])[0]
    lo = c_unsigned_int.unpack(dec_value[12:16])[0]
    digits = tuple(int(digit) for digit in f"{hi:08}{middle:08}{lo:08}".lstrip('0'))

    # Create a decimal context with the appropriate precision and exponent range
    decimal_ctx = decimal.Context(
        prec = DECIMAL128_MAX_DIGITS,
        Emin = DECIMAL128_EXPONENT_MIN,
        Emax = DECIMAL128_EXPONENT_MAX,
        capitals =  1,
        clamp = 1,
    )

    # Create the decimal value using the context
    with decimal.localcontext(decimal_ctx) as ctx:
        return buffer[16:], ctx.create_decimal((sign, digits, exponent))

def _decode_list_small(buffer: memoryview) -> Tuple[memoryview, List[Any]]:
    count = buffer[1]
    buffer = buffer[2:]
    values = [None] * count
    for i in range(count):
        buffer, values[i] = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
    return buffer, values


def _decode_list_large(buffer: memoryview) -> Tuple[memoryview, List[Any]]:
    count = c_unsigned_long.unpack(buffer[4:8])[0]
    buffer = buffer[8:]
    values = [None] * count
    for i in range(count):
        buffer, values[i] = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
    return buffer, values


def _decode_map_small(buffer: memoryview) -> Tuple[memoryview, Dict[Any, Any]]:
    count = int(buffer[1] / 2)
    buffer = buffer[2:]
    values = {}
    for _ in range(count):
        buffer, key = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
        buffer, value = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
        values[key] = value
    return buffer, values


def _decode_map_large(buffer: memoryview) -> Tuple[memoryview, Dict[Any, Any]]:
    count = int(c_unsigned_long.unpack(buffer[4:8])[0] / 2)
    buffer = buffer[8:]
    values = {}
    for _ in range(count):
        buffer, key = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
        buffer, value = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
        values[key] = value
    return buffer, values


def _decode_array_small(buffer: memoryview) -> Tuple[memoryview, List[Any]]:
    count = buffer[1]  # Ignore first byte (size) and just rely on count
    if count:
        subconstructor = buffer[2]
        buffer = buffer[3:]
        values = [None] * count
        for i in range(count):
            buffer, values[i] = _DECODE_BY_CONSTRUCTOR[subconstructor](buffer)
        return buffer, values
    return buffer[2:], []


def _decode_array_large(buffer: memoryview) -> Tuple[memoryview, List[Any]]:
    count = c_unsigned_long.unpack(buffer[4:8])[0]
    if count:
        subconstructor = buffer[8]
        buffer = buffer[9:]
        values = [None] * count
        for i in range(count):
            buffer, values[i] = _DECODE_BY_CONSTRUCTOR[subconstructor](buffer)
        return buffer, values
    return buffer[8:], []


def _decode_described(buffer: memoryview) -> Tuple[memoryview, object]:
    # TODO: to move the cursor of the buffer to the described value based on size of the
    #  descriptor without decoding descriptor value
    composite_type = buffer[0]
    buffer, descriptor = _DECODE_BY_CONSTRUCTOR[composite_type](buffer[1:])
    buffer, value = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
    try:
        composite_type = cast(int, _COMPOSITES[descriptor])
        return buffer, {composite_type: value}
    except KeyError:
        return buffer, value


def decode_payload(buffer: memoryview) -> Message:
    message: Dict[str, Union[Properties, Header, Dict, bytes, List]] = {}
    while buffer:
        # Ignore the first two bytes, they will always be the constructors for
        # described type then ulong.
        descriptor = buffer[2]
        buffer, value = _DECODE_BY_CONSTRUCTOR[buffer[3]](buffer[4:])

        if descriptor == 112:
            message["header"] = Header(*value)
        elif descriptor == 113:
            message["delivery_annotations"] = value
        elif descriptor == 114:
            message["message_annotations"] = value
        elif descriptor == 115:
            message["properties"] = Properties(*value)
        elif descriptor == 116:
            message["application_properties"] = value
        elif descriptor == 117:
            try:
                cast(List, message["data"]).append(value)
            except KeyError:
                message["data"] = [value]
        elif descriptor == 118:
            try:
                cast(List, message["sequence"]).append(value)
            except KeyError:
                message["sequence"] = [value]
        elif descriptor == 119:
            message["value"] = value
        elif descriptor == 120:
            message["footer"] = value
    # TODO: we can possibly swap out the Message construct with a TypedDict
    #  for both input and output so we get the best of both.
    # casting to TypedDict with named fields to allow for unpacking with **
    message_properties = cast("MessageDict", message)
    return Message(**message_properties)


def decode_frame(data: memoryview) -> Tuple[int, List[Any]]:
    # Ignore the first two bytes, they will always be the constructors for
    # described type then ulong.
    frame_type = data[2]
    compound_list_type = data[3]
    if compound_list_type == 0xD0:
        # list32 0xd0: data[4:8] is size, data[8:12] is count
        count = c_signed_int.unpack(data[8:12])[0]
        buffer = data[12:]
    else:
        # list8 0xc0: data[4] is size, data[5] is count
        count = data[5]
        buffer = data[6:]
    fields: List[Optional[memoryview]] = [None] * count
    for i in range(count):
        buffer, fields[i] = _DECODE_BY_CONSTRUCTOR[buffer[0]](buffer[1:])
    if frame_type == 20:
        fields.append(buffer)
    return frame_type, fields


def decode_empty_frame(header: memoryview) -> Tuple[int, bytes]:
    if header[0:4] == _HEADER_PREFIX:
        return 0, header.tobytes()
    if header[5] == 0:
        return 1, b"EMPTY"
    raise ValueError("Received unrecognized empty frame")


_DECODE_BY_CONSTRUCTOR: List[Callable] = cast(List[Callable], [None] * 256)
_DECODE_BY_CONSTRUCTOR[0] = _decode_described
_DECODE_BY_CONSTRUCTOR[64] = _decode_null
_DECODE_BY_CONSTRUCTOR[65] = _decode_true
_DECODE_BY_CONSTRUCTOR[66] = _decode_false
_DECODE_BY_CONSTRUCTOR[67] = _decode_zero
_DECODE_BY_CONSTRUCTOR[68] = _decode_zero
_DECODE_BY_CONSTRUCTOR[69] = _decode_empty
_DECODE_BY_CONSTRUCTOR[80] = _decode_ubyte
_DECODE_BY_CONSTRUCTOR[81] = _decode_byte
_DECODE_BY_CONSTRUCTOR[82] = _decode_uint_small
_DECODE_BY_CONSTRUCTOR[83] = _decode_ulong_small
_DECODE_BY_CONSTRUCTOR[84] = _decode_int_small
_DECODE_BY_CONSTRUCTOR[85] = _decode_long_small
_DECODE_BY_CONSTRUCTOR[86] = _decode_boolean
_DECODE_BY_CONSTRUCTOR[96] = _decode_ushort
_DECODE_BY_CONSTRUCTOR[97] = _decode_short
_DECODE_BY_CONSTRUCTOR[112] = _decode_uint_large
_DECODE_BY_CONSTRUCTOR[113] = _decode_int_large
_DECODE_BY_CONSTRUCTOR[114] = _decode_float
_DECODE_BY_CONSTRUCTOR[128] = _decode_ulong_large
_DECODE_BY_CONSTRUCTOR[129] = _decode_long_large
_DECODE_BY_CONSTRUCTOR[130] = _decode_double
_DECODE_BY_CONSTRUCTOR[131] = _decode_timestamp
_DECODE_BY_CONSTRUCTOR[148] = _decode_decimal128
_DECODE_BY_CONSTRUCTOR[152] = _decode_uuid
_DECODE_BY_CONSTRUCTOR[160] = _decode_binary_small
_DECODE_BY_CONSTRUCTOR[161] = _decode_binary_small
_DECODE_BY_CONSTRUCTOR[163] = _decode_binary_small
_DECODE_BY_CONSTRUCTOR[176] = _decode_binary_large
_DECODE_BY_CONSTRUCTOR[177] = _decode_binary_large
_DECODE_BY_CONSTRUCTOR[179] = _decode_binary_large
_DECODE_BY_CONSTRUCTOR[192] = _decode_list_small
_DECODE_BY_CONSTRUCTOR[193] = _decode_map_small
_DECODE_BY_CONSTRUCTOR[208] = _decode_list_large
_DECODE_BY_CONSTRUCTOR[209] = _decode_map_large
_DECODE_BY_CONSTRUCTOR[224] = _decode_array_small
_DECODE_BY_CONSTRUCTOR[240] = _decode_array_large



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

import logging
from uuid import uuid4
from datetime import datetime
from typing import Any, Optional, Tuple, Union

from .utils import utc_now, utc_from_timestamp
from .management_link import ManagementLink
from .message import Message, Properties
from .error import (
    AuthenticationException,
    ErrorCondition,
    TokenAuthFailure,
    TokenExpired,
)
from .constants import (
    CbsState,
    CbsAuthState,
    CBS_PUT_TOKEN,
    CBS_EXPIRATION,
    CBS_NAME,
    CBS_TYPE,
    CBS_OPERATION,
    ManagementExecuteOperationResult,
    ManagementOpenResult,
)

from .session import Session
from .authentication import JWTTokenAuth, SASTokenAuth

_LOGGER = logging.getLogger(__name__)


def check_expiration_and_refresh_status(expires_on: int, refresh_window: int) -> Tuple[bool, bool]:
    seconds_since_epoc = int(utc_now().timestamp())
    is_expired = seconds_since_epoc >= expires_on
    is_refresh_required = (expires_on - seconds_since_epoc) <= refresh_window
    return is_expired, is_refresh_required


def check_put_timeout_status(auth_timeout: float, token_put_time: int) -> bool:
    if auth_timeout > 0:
        return (int(utc_now().timestamp()) - token_put_time) >= auth_timeout
    return False


class CBSAuthenticator:  # pylint:disable=too-many-instance-attributes, disable=unused-argument
    def __init__(
        self, session: Session, auth: Union[JWTTokenAuth, SASTokenAuth], *, auth_timeout: float, **kwargs: Any
    ) -> None:
        self._session = session
        self._connection = self._session._connection
        self._mgmt_link: ManagementLink = self._session.create_request_response_link_pair(
            endpoint="$cbs",
            on_amqp_management_open_complete=self._on_amqp_management_open_complete,
            on_amqp_management_error=self._on_amqp_management_error,
            status_code_field=b"status-code",
            status_description_field=b"status-description",
        )

        # FIXME: probably can remove the None check as it should fail callable too
        if not auth.get_token or not callable(auth.get_token):  # type: ignore
            raise ValueError("get_token must be a callable object.")

        self._auth = auth
        self._encoding = "UTF-8"
        self._auth_timeout: float = auth_timeout
        self._token_put_time: Optional[int] = None
        self._expires_on: Optional[int] = None
        self._token: Optional[str] = None
        self._refresh_window: Optional[int] = None
        self._network_trace_params = {
            "amqpConnection": self._session._connection._container_id,
            "amqpSession": self._session.name,
            "amqpLink": "",
        }

        self._token_status_code: Optional[int] = None
        self._token_status_description: Optional[str] = None

        self.state = CbsState.CLOSED
        self.auth_state = CbsAuthState.IDLE

    def _put_token(self, token: str, token_type: str, audience: str, expires_on: Optional[datetime] = None) -> None:
        message = Message(  # type: ignore  # TODO: missing positional args header, etc.
            value=token,
            properties=Properties(message_id=uuid4()),  # type: ignore
            application_properties={
                CBS_NAME: audience,
                CBS_OPERATION: CBS_PUT_TOKEN,
                CBS_TYPE: token_type,
                CBS_EXPIRATION: expires_on,
            },
        )
        self._mgmt_link.execute_operation(
            message,
            self._on_execute_operation_complete,
            timeout=self._auth_timeout,
            operation=CBS_PUT_TOKEN,
            type=token_type,
        )

    def _on_amqp_management_open_complete(self, management_open_result: ManagementOpenResult) -> None:
        if self.state in (CbsState.CLOSED, CbsState.ERROR):
            _LOGGER.debug(
                "CSB with status: %r encounters unexpected AMQP management open complete.",
                self.state,
                extra=self._network_trace_params,
            )
        elif self.state == CbsState.OPEN:
            self.state = CbsState.ERROR
            _LOGGER.info(
                "Unexpected AMQP management open complete in OPEN, CBS error occurred.",
                extra=self._network_trace_params,
            )
        elif self.state == CbsState.OPENING:
            self.state = CbsState.OPEN if management_open_result == ManagementOpenResult.OK else CbsState.CLOSED
            _LOGGER.debug(
                "CBS completed opening with status: %r", management_open_result, extra=self._network_trace_params
            )

    def _on_amqp_management_error(self) -> None:
        if self.state == CbsState.CLOSED:
            _LOGGER.info("Unexpected AMQP error in CLOSED state.", extra=self._network_trace_params)
        elif self.state == CbsState.OPENING:
            self.state = CbsState.ERROR
            self._mgmt_link.close()
            _LOGGER.info(
                "CBS failed to open with status: %r", ManagementOpenResult.ERROR, extra=self._network_trace_params
            )
        elif self.state == CbsState.OPEN:
            self.state = CbsState.ERROR
            _LOGGER.info("CBS error occurred.", extra=self._network_trace_params)

    def _on_execute_operation_complete(
        self,
        execute_operation_result: ManagementExecuteOperationResult,
        status_code: int,
        status_description: str,
        _,
        error_condition: Optional[str] = None,
    ) -> None:
        if error_condition:
            _LOGGER.info("CBS Put token error: %r", error_condition, extra=self._network_trace_params)
            self.auth_state = CbsAuthState.ERROR
            return
        _LOGGER.debug(
            "CBS Put token result (%r), status code: %s, status_description: %s.",
            execute_operation_result,
            status_code,
            status_description,
            extra=self._network_trace_params,
        )
        self._token_status_code = status_code
        self._token_status_description = status_description

        if execute_operation_result == ManagementExecuteOperationResult.OK:
            self.auth_state = CbsAuthState.OK
        elif execute_operation_result == ManagementExecuteOperationResult.ERROR:
            self.auth_state = CbsAuthState.ERROR
            # put-token-message sending failure, rejected
            self._token_status_code = 0
            self._token_status_description = "Auth message has been rejected."
        elif execute_operation_result == ManagementExecuteOperationResult.FAILED_BAD_STATUS:
            self.auth_state = CbsAuthState.ERROR

    def _update_status(self) -> None:
        if self.auth_state in (CbsAuthState.OK, CbsAuthState.REFRESH_REQUIRED):
            is_expired, is_refresh_required = check_expiration_and_refresh_status(
                self._expires_on, self._refresh_window  # type: ignore
            )
            _LOGGER.debug(
                "CBS status check: state == %r, expired == %r, refresh required == %r",
                self.auth_state,
                is_expired,
                is_refresh_required,
                extra=self._network_trace_params,
            )
            if is_expired:
                self.auth_state = CbsAuthState.EXPIRED
            elif is_refresh_required:
                self.auth_state = CbsAuthState.REFRESH_REQUIRED
        elif self.auth_state == CbsAuthState.IN_PROGRESS:
            _LOGGER.debug(
                "CBS update in progress. Token put time: %r", self._token_put_time, extra=self._network_trace_params
            )
            if self._token_put_time is not None:
                put_timeout = check_put_timeout_status(self._auth_timeout, self._token_put_time)
                if put_timeout:
                    self.auth_state = CbsAuthState.TIMEOUT

    def _cbs_link_ready(self) -> Optional[bool]:
        if self.state == CbsState.OPEN:
            return True
        if self.state != CbsState.OPEN:
            return False
        if self.state in (CbsState.CLOSED, CbsState.ERROR):
            raise TokenAuthFailure(
                status_code=ErrorCondition.ClientError,
                status_description="CBS authentication link is in broken status, please recreate the cbs link.",
            )
        return None

    def open(self) -> None:
        self.state = CbsState.OPENING
        self._mgmt_link.open()

    def close(self) -> None:
        self._mgmt_link.close()
        self.state = CbsState.CLOSED

    def update_token(self) -> None:
        self.auth_state = CbsAuthState.IN_PROGRESS
        access_token = self._auth.get_token()
        if not access_token:
            _LOGGER.info("Token refresh function received an empty token object.", extra=self._network_trace_params)
        elif not access_token.token:
            _LOGGER.info("Token refresh function received an empty token.", extra=self._network_trace_params)
        self._expires_on = access_token.expires_on
        expires_in = self._expires_on - int(utc_now().timestamp())
        self._refresh_window = int(float(expires_in) * 0.1)
        token_type: Optional[str] = None

        if isinstance(access_token.token, bytes):
            self._token = access_token.token.decode()
        elif isinstance(access_token.token, str):
            self._token = access_token.token
        else:
            raise ValueError("Token must be a string or bytes.")
        if isinstance(self._auth.token_type, bytes):
            token_type = self._auth.token_type.decode()
        elif isinstance(self._auth.token_type, str):
            token_type = self._auth.token_type
        else:
            raise ValueError("Token type must be a string or bytes.")

        self._token_put_time = int(utc_now().timestamp())
        if self._token and token_type:
            self._put_token(
                self._token,
                token_type,
                self._auth.audience,  # type: ignore
                utc_from_timestamp(self._expires_on),
            )

    def handle_token(self) -> bool:  # pylint: disable=inconsistent-return-statements
        if not self._cbs_link_ready():
            return False
        self._update_status()
        if self.auth_state == CbsAuthState.IDLE:
            self.update_token()
            return False
        if self.auth_state == CbsAuthState.IN_PROGRESS:
            return False
        if self.auth_state == CbsAuthState.OK:
            return True
        if self.auth_state == CbsAuthState.REFRESH_REQUIRED:
            _LOGGER.info("Token will expire soon - attempting to refresh.", extra=self._network_trace_params)
            self.update_token()
            return False
        if self.auth_state == CbsAuthState.FAILURE:
            raise AuthenticationException(
                condition=ErrorCondition.InternalError,
                description="Failed to open CBS authentication link.",
            )
        if self.auth_state == CbsAuthState.ERROR:
            raise TokenAuthFailure(
                self._token_status_code,
                self._token_status_description,
                encoding=self._encoding,  # TODO: drop off all the encodings
            )
        if self.auth_state == CbsAuthState.TIMEOUT:
            raise TimeoutError("Authentication attempt timed-out.")
        if self.auth_state == CbsAuthState.EXPIRED:
            raise TokenExpired(
                condition=ErrorCondition.InternalError,
                description="CBS Authentication Expired.",
            )



# -- coding: utf-8 --
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import asyncio
import pytest
import time
import json

from azure.core.tracing import SpanKind
from azure.eventhub import EventData, TransportType, EventDataBatch
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub.exceptions import EventDataSendError, OperationTimeoutError
from azure.eventhub.amqp import (
    AmqpMessageHeader,
    AmqpMessageBodyType,
    AmqpAnnotatedMessage,
    AmqpMessageProperties,
)

try:
    import uamqp
    from uamqp.constants import TransportType as uamqp_TransportType, MessageState
    from uamqp.message import MessageProperties
except (ModuleNotFoundError, ImportError):
    uamqp = None
    uamqp_TransportType = TransportType
    MessageProperties = None
from azure.eventhub._pyamqp.message import Properties
from azure.eventhub._pyamqp.authentication import SASTokenAuth
from azure.eventhub._pyamqp.client import ReceiveClient
from azure.eventhub._pyamqp.error import AMQPConnectionError
from azure.eventhub._utils import transform_outbound_single_message

try:
    import uamqp
    from uamqp import compat
    from azure.eventhub.aio._transport._uamqp_transport_async import (
        UamqpTransportAsync as UamqpTransport,
    )
except (ModuleNotFoundError, ImportError):
    uamqp = None
    UamqpTransport = None

from azure.eventhub.aio._transport._pyamqp_transport_async import (
    PyamqpTransportAsync as PyamqpTransport,
)


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_amqp_annotated_message(auth_credentials_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        sequence_body = [b"message", 123.456, True]
        footer = {"footer_key": "footer_value"}
        prop = {"subject": "sequence"}
        seq_app_prop = {"body_type": "sequence"}

        sequence_message = AmqpAnnotatedMessage(
            sequence_body=sequence_body,
            footer=footer,
            properties=prop,
            application_properties=seq_app_prop,
        )

        value_body = {b"key": [-123, b"data", False]}
        header = {"priority": 10}
        anno = {"ann_key": "ann_value"}
        value_app_prop = {"body_type": "value"}
        value_message = AmqpAnnotatedMessage(
            value_body=value_body,
            header=header,
            annotations=anno,
            application_properties=value_app_prop,
        )

        data_body = [b"aa", b"bb", b"cc"]
        data_app_prop = {"body_type": "data"}
        del_anno = {"delann_key": "delann_value"}
        data_message = AmqpAnnotatedMessage(
            data_body=data_body,
            header=header,
            delivery_annotations=del_anno,
            application_properties=data_app_prop,
        )

        body_ed = """{"json_key": "json_val"}"""
        prop_ed = {"raw_prop": "raw_value"}
        cont_type_ed = "text/plain"
        corr_id_ed = "corr_id"
        mess_id_ed = "mess_id"
        event_data = EventData(body_ed)
        event_data.content_type = cont_type_ed
        event_data.correlation_id = corr_id_ed
        event_data.message_id = mess_id_ed

        batch = await client.create_batch()
        batch.add(data_message)
        batch.add(value_message)
        batch.add(sequence_message)
        batch.add(event_data)
        await client.send_batch(batch)
        await client.send_batch([data_message, value_message, sequence_message, event_data])
        await client.send_event(data_message)
        await client.send_event(value_message)
        await client.send_event(sequence_message)
        await client.send_event(event_data)

    received_count = {}
    received_count["data_msg"] = 0
    received_count["seq_msg"] = 0
    received_count["value_msg"] = 0
    received_count["normal_msg"] = 0

    def check_values(event):
        raw_amqp_message = event.raw_amqp_message
        if raw_amqp_message.body_type == AmqpMessageBodyType.DATA:
            if (
                raw_amqp_message.application_properties
                and raw_amqp_message.application_properties.get(b"body_type") == b"data"
            ):
                body = [data for data in raw_amqp_message.body]
                assert data_body == body
                assert event.body_as_str() == "aabbcc"
                assert raw_amqp_message.delivery_annotations[b"delann_key"] == b"delann_value"
                assert raw_amqp_message.application_properties[b"body_type"] == b"data"
                received_count["data_msg"] += 1
            else:
                assert event.body_as_json() == {"json_key": "json_val"}
                assert event.correlation_id == corr_id_ed
                assert event.message_id == mess_id_ed
                assert event.content_type == cont_type_ed
                assert event.body_type == AmqpMessageBodyType.DATA
                received_count["normal_msg"] += 1
        elif raw_amqp_message.body_type == AmqpMessageBodyType.SEQUENCE:
            body = [sequence for sequence in raw_amqp_message.body]
            assert [sequence_body] == body
            assert event.body_as_str() == "['message', 123.456, True]"
            assert raw_amqp_message.footer[b"footer_key"] == b"footer_value"
            assert raw_amqp_message.properties.subject == b"sequence"
            assert raw_amqp_message.application_properties[b"body_type"] == b"sequence"
            received_count["seq_msg"] += 1
        elif raw_amqp_message.body_type == AmqpMessageBodyType.VALUE:
            assert raw_amqp_message.body == value_body
            assert event.body_as_str() == "{'key': [-123, 'data', False]}"
            assert raw_amqp_message.annotations[b"ann_key"] == b"ann_value"
            assert raw_amqp_message.application_properties[b"body_type"] == b"value"
            received_count["value_msg"] += 1

    async def on_event(partition_context, event):
        on_event.received.append(event)

    on_event.received = []
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        task = asyncio.ensure_future(client.receive(on_event, starting_position="-1"))
        await asyncio.sleep(15)
        for event in on_event.received:
            check_values(event)

    await task

    assert len(on_event.received) == 12
    assert received_count["data_msg"] == 3
    assert received_count["seq_msg"] == 3
    assert received_count["value_msg"] == 3
    assert received_count["normal_msg"] == 3


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_partition_key_async(
    auth_credential_receivers_async, live_eventhub, uamqp_transport, timeout_factor, client_args
):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        data_val = 0
        for partition in [b"a", b"b", b"c", b"d", b"e", b"f"]:
            partition_key = b"test_partition_" + partition
            for i in range(10):
                batch = await client.create_batch(partition_key=partition_key)
                batch.add(EventData(str(data_val)))
                data_val += 1
                await client.send_batch(batch)

        await client.send_batch(await client.create_batch())

        for partition in [b"a", b"b", b"c", b"d", b"e", b"f"]:
            partition_key = b"test_partition_" + partition
            for i in range(10):
                event_data = EventData(str(data_val))
                event_data.properties = {"is_single": True}
                data_val += 1
                await client.send_event(event_data, partition_key=partition_key)

    batch_cnt = 0
    single_cnt = 0
    found_partition_keys = {}
    for index, partition in enumerate(receivers):
        retry_total = 0
        while retry_total < 3:
            timeout = (5 * retry_total) * timeout_factor
            received = partition.receive_message_batch(timeout=timeout)
            for message in received:
                try:
                    event_data = EventData._from_message(message)
                    if event_data.properties and event_data.properties[b"is_single"]:
                        single_cnt += 1
                    else:
                        batch_cnt += 1
                    existing = found_partition_keys[event_data.partition_key]
                    assert existing == index
                except KeyError:
                    found_partition_keys[event_data.partition_key] = index
            if received:
                break
            retry_total += 1
        if retry_total == 3:
            raise OperationTimeoutError(f"Exhausted retries for receiving from {live_eventhub['hostname']}.")

    assert single_cnt == 60
    assert batch_cnt == 60
    assert len(found_partition_keys) == 6


@pytest.mark.parametrize("payload", [b"", b"A single event"])
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_and_receive_small_body_async(
    auth_credential_receivers_async, payload, uamqp_transport, timeout_factor, client_args
):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async

    # TODO: Commenting out tracing for now. Need to fix this issue first: #36571
    # fake_span = enable_tracing
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    # with fake_span(name="SendTest") as root_span:
    async with client:
        batch = await client.create_batch()
        batch.add(EventData(payload))
        batch.add(EventData(payload))

        await client.send_batch(batch)
        await client.send_event(EventData(payload))
    received = []
    for r in receivers:
        received.extend([EventData._from_message(x) for x in r.receive_message_batch(timeout=5 * timeout_factor)])

    assert len(received) == 3
    assert list(received[0].body)[0] == payload
    assert list(received[1].body)[0] == payload
    assert list(received[2].body)[0] == payload

    ## Will need to modify FakeSpan in conftest.
    # assert root_span.name == "SendTest"
    # assert len(root_span.children) == 5

    ## Check first message added to batch.
    # assert root_span.children[0].name == "EventHubs.message"
    # assert root_span.children[0].kind == SpanKind.PRODUCER

    ## Check second message added to batch.
    # assert root_span.children[1].name == "EventHubs.message"
    # assert root_span.children[1].kind == SpanKind.PRODUCER

    ## Check send span corresponding to send_batch
    # assert root_span.children[2].name == "EventHubs.send"
    # assert root_span.children[2].kind == SpanKind.CLIENT
    # assert len(root_span.children[2].links) == 2
    # assert (
    #    root_span.children[2].links[0].headers["traceparent"]
    #    == root_span.children[0].traceparent
    # )
    # assert (
    #    root_span.children[2].links[1].headers["traceparent"]
    #    == root_span.children[1].traceparent
    # )

    ## Check message sent using send_event
    # assert root_span.children[3].name == "EventHubs.message"
    # assert root_span.children[3].kind == SpanKind.PRODUCER

    ## Check send span corresponding to send_event
    # assert root_span.children[4].name == "EventHubs.send"
    # assert root_span.children[4].kind == SpanKind.CLIENT
    # assert len(root_span.children[4].links) == 1
    # assert (
    #    root_span.children[4].links[0].headers["traceparent"]
    #    == root_span.children[3].traceparent
    # )


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_partition_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with client:
        batch = await client.create_batch()
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        await client.send_event(EventData(b"Data"))

    async with client:
        batch = await client.create_batch(partition_id="1")
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        await client.send_event(EventData(b"Data"), partition_id="1")

    partition_0 = receivers[0].receive_message_batch(timeout=10 * timeout_factor)
    partition_1 = receivers[1].receive_message_batch(timeout=10 * timeout_factor)
    assert len(partition_1) >= 2
    assert len(partition_0) + len(partition_1) == 4

    async with client:
        batch = await client.create_batch()
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        await client.send_event(EventData(b"Data"))
    async with client:
        batch = await client.create_batch(partition_id="0")
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        await client.send_event(EventData(b"Data"), partition_id="0")
    async with client:
        batch = EventDataBatch(partition_id="0")
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)

    time.sleep(5)
    partition_0 = receivers[0].receive_message_batch(timeout=10 * timeout_factor)
    partition_1 = receivers[1].receive_message_batch(timeout=10 * timeout_factor)
    assert len(partition_0) >= 3
    assert len(partition_0) + len(partition_1) == 5


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_non_ascii_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        batch = await client.create_batch(partition_id="0")
        batch.add(EventData("é,è,à,ù,â,ê,î,ô,û"))
        batch.add(EventData(json.dumps({"foo": "漢字"})))
        await client.send_batch(batch)
        await client.send_event(EventData("é,è,à,ù,â,ê,î,ô,û"), partition_id="0")
        await client.send_event(EventData(json.dumps({"foo": "漢字"})), partition_id="0")
    await asyncio.sleep(1)
    # receive_message_batch() returns immediately once it receives any messages before the max_batch_size
    # and timeout reach. Could be 1, 2, or any number between 1 and max_batch_size.
    # So call it twice to ensure the two events are received.
    partition_0 = [
        EventData._from_message(x) for x in receivers[0].receive_message_batch(timeout=5 * timeout_factor)
    ] + [EventData._from_message(x) for x in receivers[0].receive_message_batch(timeout=5 * timeout_factor)]

    assert len(partition_0) == 4
    assert partition_0[0].body_as_str() == "é,è,à,ù,â,ê,î,ô,û"
    assert partition_0[1].body_as_json() == {"foo": "漢字"}
    assert partition_0[2].body_as_str() == "é,è,à,ù,â,ê,î,ô,û"
    assert partition_0[3].body_as_json() == {"foo": "漢字"}


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_multiple_partition_with_app_prop_async(
    auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args
):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    app_prop_key = "raw_prop"
    app_prop_value = "raw_value"
    app_prop = {app_prop_key: app_prop_value}
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        ed0 = EventData(b"Message 0")
        ed0.properties = app_prop
        batch = await client.create_batch(partition_id="0")
        batch.add(ed0)
        await client.send_batch(batch)
        await client.send_event(ed0, partition_id="0")

        ed1 = EventData(b"Message 1")
        ed1.properties = app_prop
        batch = await client.create_batch(partition_id="1")
        batch.add(ed1)
        await client.send_batch(batch)
        await client.send_event(ed1, partition_id="1")
    partition_0 = [EventData._from_message(x) for x in receivers[0].receive_message_batch(timeout=5 * timeout_factor)]
    assert len(partition_0) == 2
    assert partition_0[0].properties[b"raw_prop"] == b"raw_value"
    assert partition_0[1].properties[b"raw_prop"] == b"raw_value"
    partition_1 = [EventData._from_message(x) for x in receivers[1].receive_message_batch(timeout=5 * timeout_factor)]
    assert len(partition_0) == 2
    assert partition_1[0].properties[b"raw_prop"] == b"raw_value"
    assert partition_0[1].properties[b"raw_prop"] == b"raw_value"


@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.no_amqpproxy # Proxy requires TransportType.Amqp
async def test_send_over_websocket_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        transport_type=uamqp_TransportType.AmqpOverWebsocket,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with client:
        batch = await client.create_batch(partition_id="0")
        batch.add(EventData("Event Data"))
        await client.send_batch(batch)
        await client.send_event(EventData("Event Data"), partition_id="0")

    time.sleep(1)
    received = []
    received.extend(receivers[0].receive_message_batch(max_batch_size=5, timeout=10 * timeout_factor))
    assert len(received) == 2


@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.no_amqpproxy # Proxy requires TransportType.Amqp
async def test_send_with_create_event_batch_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    app_prop_key = "raw_prop"
    app_prop_value = "raw_value"
    app_prop = {app_prop_key: app_prop_value}
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        transport_type=TransportType.AmqpOverWebsocket,
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        event_data_batch = await client.create_batch(max_size_in_bytes=100000)
        while True:
            try:
                ed = EventData("A single event data")
                ed.properties = app_prop
                event_data_batch.add(ed)
            except ValueError:
                break
        await client.send_batch(event_data_batch)
        received = []
        for r in receivers:
            received.extend(r.receive_message_batch(timeout=10 * timeout_factor))
        assert len(received) >= 1
        assert EventData._from_message(received[0]).properties[b"raw_prop"] == b"raw_value"


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_list_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    payload = "A1"
    async with client:
        await client.send_batch([EventData(payload)])
    received = []
    for r in receivers:
        received.extend([EventData._from_message(x) for x in r.receive_message_batch(timeout=10 * timeout_factor)])

    assert len(received) == 1
    assert received[0].body_as_str() == payload


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_list_partition_async(auth_credential_receivers_async, uamqp_transport, timeout_factor, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    payload = "A1"
    async with client:
        await client.send_batch([EventData(payload)], partition_id="0")
        message = receivers[0].receive_message_batch(timeout=10 * timeout_factor)[0]
        received = EventData._from_message(message)
    assert received.body_as_str() == payload


@pytest.mark.parametrize(
    "to_send, exception_type",
    [([EventData("A" * 1024)] * 1100, ValueError), ("any str", AttributeError)],
)
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_list_wrong_data_async(auth_credentials_async, to_send, exception_type, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args
    )
    async with client:
        with pytest.raises(exception_type):
            await client.send_batch(to_send)


@pytest.mark.parametrize("partition_id, partition_key", [("0", None), (None, "pk")])
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_batch_pid_pk_async(invalid_hostname, partition_id, partition_key, uamqp_transport, client_args):
    # Use invalid_hostname because this is not a live test.
    client = EventHubProducerClient.from_connection_string(invalid_hostname, uamqp_transport=uamqp_transport, **client_args)
    batch = EventDataBatch(partition_id=partition_id, partition_key=partition_key)
    async with client:
        with pytest.raises(TypeError):
            await client.send_batch(batch, partition_id=partition_id, partition_key=partition_key)


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_callback_async(auth_credentials_async, uamqp_transport, client_args):

    async def on_error(events, pid, err):
        on_error.err = err

    async def on_success(events, pid):
        sent_events.append((events, pid))

    sent_events = []
    on_error.err = None
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with client:
        batch = await client.create_batch()
        batch.add(EventData(b"Data"))
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        assert len(sent_events[-1][0]) == 2
        assert not sent_events[-1][1]
        await client.send_event(EventData(b"Data"))
        assert len(sent_events[-1][0]) == 1
        assert not sent_events[-1][1]

        batch = await client.create_batch(partition_key="key")
        batch.add(EventData(b"Data"))
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        assert len(sent_events[-1][0]) == 2
        assert not sent_events[-1][1]
        await client.send_event(EventData(b"Data"), partition_key="key")
        assert len(sent_events[-1][0]) == 1
        assert not sent_events[-1][1]

        batch = await client.create_batch(partition_id="0")
        batch.add(EventData(b"Data"))
        await client.send_batch(batch)
        batch.add(EventData(b"Data"))
        assert len(sent_events[-1][0]) == 2
        assert sent_events[-1][1] == "0"
        await client.send_event(EventData(b"Data"), partition_id="0")
        assert len(sent_events[-1][0]) == 1
        assert sent_events[-1][1] == "0"

        assert not on_error.err


@pytest.mark.parametrize("keep_alive", [None, 30, 60])
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_keep_alive_async(auth_credentials_async, keep_alive, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        keep_alive=keep_alive,
        uamqp_transport=uamqp_transport,
        **client_args
    )
    assert client._producers["all-partitions"]._keep_alive == keep_alive


# TODO: This test fails both sync and async with amqp proxy and keep_alive=30. Investigate.
@pytest.mark.parametrize("keep_alive", [None, 5, 30])
@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.no_amqpproxy
async def test_send_long_wait_idle_timeout(auth_credentials_async, keep_alive, uamqp_transport, client_args):
    if uamqp_transport:
        amqp_transport = UamqpTransport
        retry_total = 3
    else:
        amqp_transport = PyamqpTransport
        retry_total = 0
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        keep_alive=keep_alive,
        idle_timeout=10,
        retry_total=retry_total,
        uamqp_transport=uamqp_transport,
        **client_args
    )
    sender = client._create_producer(partition_id="0")
    async with sender:
        await sender._open_with_retry()
        ed = EventData("data")
        ed = transform_outbound_single_message(ed, EventData, amqp_transport.to_outgoing_amqp_message)
        sender._unsent_events = [ed._message]
        # hit idle timeout error
        await asyncio.sleep(11)

        if uamqp_transport:
            sender._unsent_events[0].on_send_complete = sender._on_outcome
            if keep_alive != 5:
                with pytest.raises(
                    (
                        uamqp.errors.ConnectionClose,
                        uamqp.errors.MessageHandlerError,
                        OperationTimeoutError,
                    )
                ):
                    await sender._send_event_data()
            else:
                await sender._send_event_data()

        if not uamqp_transport:
            if keep_alive == 5:
                await sender._send_event_data()
            else:
                with pytest.raises(AMQPConnectionError):
                    await sender._send_event_data()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from enum import Enum


class CloseReason(Enum):
    """The reason a partition consumer is closed."""

    # The Consumer was explicitly required to stop. This would happen when the EventHubConsumerClient is closed.
    SHUTDOWN = 0

    # The Consumer lost the ownership of a partition. This would happend when another EventHubConsumerClient
    # instance claims ownership of the partition.
    OWNERSHIP_LOST = 1


class LoadBalancingStrategy(Enum):
    GREEDY = "greedy"
    BALANCED = "balanced"



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to show sending, receiving and parsing amqp annotated message(s) to Event Hubs.
"""

import os
from azure.eventhub import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub.amqp import AmqpAnnotatedMessage, AmqpMessageBodyType
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


def send_data_message(producer):
    data_body = [b"aa", b"bb", b"cc"]
    application_properties = {"body_type": "data"}
    delivery_annotations = {"delivery_annotation_key": "value"}
    data_message = AmqpAnnotatedMessage(
        data_body=data_body, delivery_annotations=delivery_annotations, application_properties=application_properties
    )
    batch = producer.create_batch()
    batch.add(data_message)
    producer.send_batch(batch)
    print("Message of data body sent.")


def send_sequence_message(producer):
    sequence_body = [b"message", 123.456, True]
    footer = {"footer_key": "footer_value"}
    properties = {"subject": "sequence"}
    application_properties = {"body_type": "sequence"}
    sequence_message = AmqpAnnotatedMessage(
        sequence_body=sequence_body, footer=footer, properties=properties, application_properties=application_properties
    )
    producer.send_batch([sequence_message])
    print("Message of sequence body sent.")


def send_value_message(producer):
    value_body = {b"key": [-123, b"data", False]}
    header = {"priority": 10}
    annotations = {"annotation_key": "value"}
    application_properties = {"body_type": "value"}
    value_message = AmqpAnnotatedMessage(
        value_body=value_body, header=header, annotations=annotations, application_properties=application_properties
    )
    producer.send_batch([value_message])
    print("Message of value body sent.")


def on_event(partition_context, event):
    # Put your code here.
    # If the operation is i/o intensive, multi-thread will have better performance.
    print("Received event from partition: {}".format(partition_context.partition_id))
    raw_amqp_message = event.raw_amqp_message
    if raw_amqp_message.body_type == AmqpMessageBodyType.DATA:
        print("Message of data body received. Body is:")
        for data_section in raw_amqp_message.body:
            print(data_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.SEQUENCE:
        print("Message of sequence body received. Body is:")
        for sequence_section in raw_amqp_message.body:
            print(sequence_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.VALUE:
        print("Message of value body received. Body is:")
        print(raw_amqp_message.body)


def receive_and_parse_message(consumer):
    with consumer:
        try:
            consumer.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        except KeyboardInterrupt:
            print("Stopped receiving.")


producer = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
)
with producer:
    send_data_message(producer)
    send_sequence_message(producer)
    send_value_message(producer)


consumer = EventHubConsumerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
    consumer_group="$Default",
)
receive_and_parse_message(consumer)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import pytest

from azure.eventhub._pyamqp import SendClient, Connection, authentication
from azure.eventhub._pyamqp.error import AMQPConnectionError


def test_client_creation_exceptions():
    with pytest.raises(TypeError):
        sender = SendClient(
            "fake.host.com",
        )
        assert sender._hostname == "fake.host.com"


def test_connection_endpoint_exceptions():
    with pytest.raises(AMQPConnectionError):
        endpoint = "sb://fake.host.com"
        connection = Connection(endpoint)
        connection.open()


def test_connection_sas_authentication_exception():
    uri = "sb://{}/{}".format("fake.host.come", "fake_eh")

    target = "amqps://{}/{}/Partitions/{}".format("fake.host.com", "fake_eh", "0")
    sas_auth = authentication.SASTokenAuth(uri=uri, audience=uri, username="key", password="")
    with pytest.raises(AttributeError):
        sender = SendClient("fake.host.com", target, auth=sas_auth)
        sender.client_ready()


def test_connection_sasl_annon_authentication_exception():
    target = "amqps://{}/{}/Partitions/{}".format("fake.host.com", "fake_eh", "0")

    sas_auth = authentication.SASLAnonymousCredential()
    with pytest.raises(AttributeError):
        sender = SendClient("fake.host.com", target, auth=sas_auth)
        sender.client_ready()



import time
import pytest
import threading
import sys

from azure.eventhub import EventData
from azure.eventhub import EventHubConsumerClient
from azure.eventhub._eventprocessor.in_memory_checkpoint_store import (
    InMemoryCheckpointStore,
)
from azure.eventhub._constants import ALL_PARTITIONS


@pytest.mark.liveTest
def test_receive_storage_checkpoint(
    auth_credential_senders,
    uamqp_transport,
    client_args,
    checkpoint_store,
    live_eventhub,
    resource_mgmt_client,
):
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders

    for i in range(10):
        senders[0].send(EventData("Test EventData"))
        senders[1].send(EventData("Test EventData"))

    try:
        checkpoint_store._container_client.create_container()
    except:
        pass

    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        checkpoint_store=checkpoint_store,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    sequence_numbers_0 = []
    sequence_numbers_1 = []

    def on_event(partition_context, event):
        partition_context.update_checkpoint(event)
        sequence_num = event.sequence_number
        if partition_context.partition_id == "0":
            if sequence_num in sequence_numbers_0:
                assert False
            sequence_numbers_0.append(sequence_num)
        else:
            if sequence_num in sequence_numbers_1:
                assert False
            sequence_numbers_1.append(sequence_num)

    with client:
        worker = threading.Thread(target=client.receive, args=(on_event,), kwargs={"starting_position": "-1"})
        worker.start()

        # Update the eventhub
        eventhub = resource_mgmt_client.event_hubs.get(
            live_eventhub["resource_group"],
            live_eventhub["namespace"],
            live_eventhub["event_hub"],
        )
        properties = eventhub.as_dict()
        if properties["message_retention_in_days"] == 1:
            properties["message_retention_in_days"] = 2
        else:
            properties["message_retention_in_days"] = 1
        resource_mgmt_client.event_hubs.create_or_update(
            live_eventhub["resource_group"],
            live_eventhub["namespace"],
            live_eventhub["event_hub"],
            properties,
        )

        time.sleep(20)

    assert len(sequence_numbers_0) == 10
    assert len(sequence_numbers_1) == 10

    try:
        checkpoint_store._container_client.delete_container()
    except:
        pass


@pytest.mark.liveTest
def test_receive_no_partition(auth_credential_senders, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    senders[0].send(EventData("Test EventData"))
    senders[1].send(EventData("Test EventData"))
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        receive_timeout=1,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    def on_event(partition_context, event):
        on_event.received += 1
        partition_context.update_checkpoint(event)
        on_event.namespace = partition_context.fully_qualified_namespace
        on_event.eventhub_name = partition_context.eventhub_name
        on_event.consumer_group = partition_context.consumer_group
        on_event.offset = event.offset
        on_event.sequence_number = event.sequence_number

    on_event.received = 0
    on_event.namespace = None
    on_event.eventhub_name = None
    on_event.consumer_group = None
    on_event.offset = None
    on_event.sequence_number = None

    with client:
        worker = threading.Thread(target=client.receive, args=(on_event,), kwargs={"starting_position": "-1"})
        worker.start()
        time.sleep(20)
        assert on_event.received == 2
        checkpoints = list(client._event_processors.values())[0]._checkpoint_store.list_checkpoints(
            on_event.namespace, on_event.eventhub_name, on_event.consumer_group
        )
        assert len([checkpoint for checkpoint in checkpoints if checkpoint["offset"] == on_event.offset]) > 0
        assert (
            len([checkpoint for checkpoint in checkpoints if checkpoint["sequence_number"] == on_event.sequence_number])
            > 0
        )


@pytest.mark.liveTest
def test_receive_partition(auth_credential_senders, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    senders[0].send(EventData("Test EventData"))
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    def on_event(partition_context, event):
        on_event.received += 1
        on_event.partition_id = partition_context.partition_id
        on_event.consumer_group = partition_context.consumer_group
        on_event.fully_qualified_namespace = partition_context.fully_qualified_namespace
        on_event.eventhub_name = partition_context.eventhub_name

    on_event.received = 0
    with client:
        worker = threading.Thread(
            target=client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "partition_id": "0"},
        )
        worker.start()
        time.sleep(10)
        assert on_event.received == 1
        assert on_event.partition_id == "0"
        assert on_event.consumer_group == "$default"
        assert on_event.fully_qualified_namespace == fully_qualified_namespace
        assert on_event.eventhub_name == senders[0]._client.eventhub_name


@pytest.mark.liveTest
def test_receive_load_balancing(auth_credential_senders, uamqp_transport, client_args):
    if sys.platform.startswith("darwin"):
        pytest.skip("Skipping on OSX - test code using multiple threads. Sometimes OSX aborts python process")

    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    cs = InMemoryCheckpointStore()
    client1 = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        checkpoint_store=cs,
        load_balancing_interval=1,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    client2 = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        checkpoint_store=cs,
        load_balancing_interval=1,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    def on_event(partition_context, event):
        pass

    with client1, client2:
        worker1 = threading.Thread(target=client1.receive, args=(on_event,), kwargs={"starting_position": "-1"})

        worker2 = threading.Thread(target=client2.receive, args=(on_event,), kwargs={"starting_position": "-1"})

        worker1.start()
        time.sleep(3.3)
        worker2.start()
        time.sleep(20)
        assert len(client1._event_processors[("$default", ALL_PARTITIONS)]._consumers) == 1
        assert len(client2._event_processors[("$default", ALL_PARTITIONS)]._consumers) == 1


def test_receive_batch_no_max_wait_time(auth_credential_senders, uamqp_transport, client_args):
    """Test whether callback is called when max_wait_time is None and max_batch_size has reached"""
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    senders[0].send(EventData("Test EventData"))
    senders[1].send(EventData("Test EventData"))
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)
        partition_context.update_checkpoint()
        on_event_batch.namespace = partition_context.fully_qualified_namespace
        on_event_batch.eventhub_name = partition_context.eventhub_name
        on_event_batch.consumer_group = partition_context.consumer_group
        on_event_batch.offset = event_batch[-1].offset
        on_event_batch.sequence_number = event_batch[-1].sequence_number

    on_event_batch.received = 0
    on_event_batch.namespace = None
    on_event_batch.eventhub_name = None
    on_event_batch.consumer_group = None
    on_event_batch.offset = None
    on_event_batch.sequence_number = None

    with client:
        worker = threading.Thread(
            target=client.receive_batch,
            args=(on_event_batch,),
            kwargs={"starting_position": "-1"},
        )
        worker.start()
        time.sleep(20)
        assert on_event_batch.received == 2

        checkpoints = list(client._event_processors.values())[0]._checkpoint_store.list_checkpoints(
            on_event_batch.namespace,
            on_event_batch.eventhub_name,
            on_event_batch.consumer_group,
        )
        assert len([checkpoint for checkpoint in checkpoints if checkpoint["offset"] == on_event_batch.offset]) > 0
        assert (
            len(
                [
                    checkpoint
                    for checkpoint in checkpoints
                    if checkpoint["sequence_number"] == on_event_batch.sequence_number
                ]
            )
            > 0
        )

    worker.join()


@pytest.mark.parametrize("max_wait_time, sleep_time, expected_result", [(3, 15, []), (3, 2, None)])
def test_receive_batch_empty_with_max_wait_time(
    auth_credentials, uamqp_transport, client_args, max_wait_time, sleep_time, expected_result
):
    """Test whether event handler is called when max_wait_time > 0 and no event is received"""
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    def on_event_batch(partition_context, event_batch):
        on_event_batch.event_batch = event_batch

    on_event_batch.event_batch = None
    with client:
        worker = threading.Thread(
            target=client.receive_batch,
            args=(on_event_batch,),
            kwargs={"max_wait_time": max_wait_time, "starting_position": "-1"},
        )
        worker.start()
        time.sleep(sleep_time)
        assert on_event_batch.event_batch == expected_result
    worker.join()


def test_receive_batch_early_callback(auth_credential_senders, uamqp_transport, client_args):
    """Test whether the callback is called once max_batch_size reaches and before max_wait_time reaches."""
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    for _ in range(10):
        senders[0].send(EventData("Test EventData"))
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)

    on_event_batch.received = 0

    with client:
        worker = threading.Thread(
            target=client.receive_batch,
            args=(on_event_batch,),
            kwargs={
                "max_batch_size": 10,
                "max_wait_time": 100,
                "starting_position": "-1",
                "partition_id": "0",
            },
        )
        worker.start()
        time.sleep(10)
        assert on_event_batch.received == 10
    worker.join()


@pytest.mark.liveTest
def test_receive_batch_tracing(auth_credential_senders, uamqp_transport, client_args):
    """Test that that receive and process spans are properly created and linked."""

    # TODO: Commenting out tracing for now. Need to fix this issue first: #36571
    # fake_span = enable_tracing
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders

    # with fake_span(name="SendSpan") as root_send:
    senders[0].send([EventData(b"Data"), EventData(b"Data")])

    # assert len(root_send.children) == 3
    # assert root_send.children[0].name == "EventHubs.message"
    # assert root_send.children[1].name == "EventHubs.message"
    # assert root_send.children[2].name == "EventHubs.send"
    # assert len(root_send.children[2].links) == 2

    # traceparent1 = root_send.children[2].links[0].headers["traceparent"]
    # traceparent2 = root_send.children[2].links[1].headers["traceparent"]

    def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)

    on_event_batch.received = 0

    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    # with fake_span(name="ReceiveSpan") as root_receive:
    with client:
        worker = threading.Thread(
            target=client.receive_batch,
            args=(on_event_batch,),
            kwargs={"starting_position": "-1"},
        )
        worker.start()
        time.sleep(20)
        assert on_event_batch.received == 2

    worker.join()

    # assert root_receive.name == "ReceiveSpan"
    ## One receive span and one process span.
    # assert len(root_receive.children) == 2

    # assert root_receive.children[0].name == "EventHubs.receive"
    # assert root_receive.children[0].kind == SpanKind.CLIENT

    ## One link for each message in the batch.
    # assert len(root_receive.children[0].links) == 2
    # assert root_receive.children[0].links[0].headers["traceparent"] == traceparent1
    # assert root_receive.children[0].links[1].headers["traceparent"] == traceparent2

    # assert root_receive.children[1].name == "EventHubs.process"
    # assert root_receive.children[1].kind == SpanKind.CONSUMER

    # assert len(root_receive.children[1].links) == 2
    # assert root_receive.children[1].links[0].headers["traceparent"] == traceparent1
    # assert root_receive.children[1].links[1].headers["traceparent"] == traceparent2


@pytest.mark.liveTest
def test_receive_batch_large_event(auth_credential_senders, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential, senders = auth_credential_senders
    senders[0].send(EventData("A" * 15700))
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    def on_event(partition_context, event):
        on_event.received += 1
        on_event.partition_id = partition_context.partition_id
        on_event.consumer_group = partition_context.consumer_group
        on_event.fully_qualified_namespace = partition_context.fully_qualified_namespace
        on_event.eventhub_name = partition_context.eventhub_name
        assert client._event_processors[0]._consumers[0]._handler._link.current_link_credit == 1

    on_event.received = 0
    with client:
        worker = threading.Thread(
            target=client.receive_batch,
            args=(on_event,),
            kwargs={"starting_position": "-1", "partition_id": "0", "prefetch": 2},
        )
        worker.start()
        time.sleep(10)
        assert on_event.received == 1
        assert on_event.partition_id == "0"
        assert on_event.consumer_group == "$default"
        assert on_event.fully_qualified_namespace == fully_qualified_namespace
        assert on_event.eventhub_name == senders[0]._client.eventhub_name



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._buffered_producer import BufferedProducer
from ._partition_resolver import PartitionResolver
from ._buffered_producer_dispatcher import BufferedProducerDispatcher

__all__ = [
    "BufferedProducer",
    "PartitionResolver",
    "BufferedProducerDispatcher",
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
# TODO: Check types of kwargs (issue exists for this)
import asyncio # pylint:disable=do-not-import-asyncio
import logging
import time
import queue
from functools import partial
from typing import Any, Callable, Coroutine, List, Dict, Optional, Tuple, Union, overload, cast
from typing_extensions import Literal

from ..outcomes import Accepted, Modified, Received, Rejected, Released
from ._connection_async import Connection
from ._management_operation_async import ManagementOperation
from ._cbs_async import CBSAuthenticator
from ..client import (
    AMQPClient as AMQPClientSync,
    ReceiveClient as ReceiveClientSync,
    SendClient as SendClientSync,
    Outcomes,
)
from ..message import _MessageDelivery, Message
from ..constants import (
    MessageDeliveryState,
    SEND_DISPOSITION_ACCEPT,
    SEND_DISPOSITION_REJECT,
    LinkDeliverySettleReason,
    MESSAGE_DELIVERY_DONE_STATES,
    AUTH_TYPE_CBS,
)
from ..error import AMQPError, ErrorCondition, AMQPException, MessageException

_logger = logging.getLogger(__name__)


class AMQPClientAsync(AMQPClientSync):
    """An asynchronous AMQP client.

    :param hostname: The AMQP endpoint to connect to.
    :type hostname: str
    :keyword auth: Authentication for the connection. This should be one of the following:
        - pyamqp.authentication.SASLAnonymous
        - pyamqp.authentication.SASLPlain
        - pyamqp.authentication.SASTokenAuth
        - pyamqp.authentication.JWTTokenAuth
     If no authentication is supplied, SASLAnnoymous will be used by default.
    :paramtype auth: ~pyamqp.authentication
    :keyword client_name: The name for the client, also known as the Container ID.
     If no name is provided, a random GUID will be used.
    :paramtype client_name: str or bytes
    :keyword network_trace: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :paramtype network_trace: bool
    :keyword retry_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :paramtype retry_policy: ~pyamqp.error.RetryPolicy
    :keyword keep_alive_interval: If set, a thread will be started to keep the connection
     alive during periods of user inactivity. The value will determine how long the
     thread will sleep (in seconds) between pinging the connection. If 0 or None, no
     thread will be started.
    :paramtype keep_alive_interval: int
    :keyword max_frame_size: Maximum AMQP frame size. Default is 63488 bytes.
    :paramtype max_frame_size: int
    :keyword channel_max: Maximum number of Session channels in the Connection.
    :paramtype channel_max: int
    :keyword idle_timeout: Timeout in seconds after which the Connection will close
     if there is no further activity.
    :paramtype idle_timeout: int
    :keyword auth_timeout: Timeout in seconds for CBS authentication. Otherwise this value will be ignored.
     Default value is 60s.
    :paramtype auth_timeout: int
    :keyword properties: Connection properties.
    :paramtype properties: dict[str, any]
    :keyword remote_idle_timeout_empty_frame_send_ratio: Ratio of empty frames to
     idle time for Connections with no activity. Value must be between
     0.0 and 1.0 inclusive. Default is 0.5.
    :paramtype remote_idle_timeout_empty_frame_send_ratio: float
    :keyword incoming_window: The size of the allowed window for incoming messages.
    :paramtype incoming_window: int
    :keyword outgoing_window: The size of the allowed window for outgoing messages.
    :paramtype outgoing_window: int
    :keyword handle_max: The maximum number of concurrent link handles.
    :paramtype handle_max: int
    :keyword on_attach: A callback function to be run on receipt of an ATTACH frame.
     The function must take 4 arguments: source, target, properties and error.
    :paramtype on_attach: func[
     ~pyamqp.endpoint.Source, ~pyamqp.endpoint.Target, dict, ~pyamqp.error.AMQPConnectionError]
    :keyword send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :paramtype send_settle_mode: ~pyamqp.constants.SenderSettleMode
    :keyword receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :paramtype receive_settle_mode: ~pyamqp.constants.ReceiverSettleMode
    :keyword desired_capabilities: The extension capabilities desired from the peer endpoint.
    :paramtype desired_capabilities: list[bytes]
    :keyword max_message_size: The maximum allowed message size negotiated for the Link.
    :paramtype max_message_size: int
    :keyword link_properties: Metadata to be sent in the Link ATTACH frame.
    :paramtype link_properties: dict[str, any]
    :keyword link_credit: The Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
     The default is 300.
    :paramtype link_credit: int
    :keyword transport_type: The type of transport protocol that will be used for communicating with
     the service. Default is `TransportType.Amqp` in which case port 5671 is used.
     If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
     be used instead which uses port 443 for communication.
    :paramtype transport_type: ~pyamqp.constants.TransportType
    :keyword http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
     Additionally the following keys may also be present: `'username', 'password'`.
    :paramtype http_proxy: dict[str, str]
    :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
     the Event Hubs service, allowing network requests to be routed through any application gateways or
     other paths needed for the host environment. Default is None.
     Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
     If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
    :paramtype custom_endpoint_address: str
    :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
     authenticate the identity of the connection endpoint. Ignored if ssl_context passed in. Default is None
     in which case `certifi.where()` will be used.
    :paramtype connection_verify: str
    :keyword ssl_context: An instance of ssl.SSLContext to be used. If this is specified, connection_verify is ignored.
    :paramtype ssl_context: ssl.SSLContext or None
    :keyword float socket_timeout: The maximum time in seconds that the underlying socket in the transport should
     wait when reading or writing data before timing out. The default value is 0.2 (for transport type Amqp),
     and 1 for transport type AmqpOverWebsocket.
    """

    def __init__(self, hostname, **kwargs):
        self._mgmt_link_lock_async = asyncio.Lock()
        super().__init__(hostname, **kwargs)

    async def _keep_alive_async(self):
        start_time = time.time()
        try:
            while self._connection and not self._shutdown:
                current_time = time.time()
                elapsed_time = current_time - start_time
                if elapsed_time >= self._keep_alive_interval:
                    await asyncio.shield(
                        self._connection.listen(wait=self._socket_timeout, batch=self._link.current_link_credit)
                    )
                    start_time = current_time
                await asyncio.sleep(1)
        except Exception as e:  # pylint: disable=broad-except
            _logger.info(
                "Connection keep-alive for %r failed: %r.", self.__class__.__name__, e, extra=self._network_trace_params
            )

    async def __aenter__(self):
        """Run Client in an async context manager.
        :return: The Client object.
        :rtype: ~pyamqp.AMQPClient
        """
        await self.open_async()
        return self

    async def __aexit__(self, *args):
        """Close and destroy Client on exiting an async context manager.
        :param any args: Ignored.
        """
        await self.close_async()

    async def _client_ready_async(self):
        """Determine whether the client is ready to start sending and/or
        receiving messages. To be ready, the connection must be open and
        authentication complete.

        :return: Whether or not the client is ready for operation.
        :rtype: bool
        """
        return True

    async def _client_run_async(self, **kwargs):
        """Perform a single Connection iteration."""
        await self._connection.listen(wait=self._socket_timeout, **kwargs)

    async def _close_link_async(self):
        if self._link and not self._link._is_closed:  # pylint: disable=protected-access
            await self._link.detach(close=True)
            self._link = None

    async def _do_retryable_operation_async(self, operation, *args, **kwargs):
        retry_settings = self._retry_policy.configure_retries()
        retry_active = True
        absolute_timeout = kwargs.pop("timeout", 0) or 0
        start_time = time.time()
        while retry_active:
            try:
                if absolute_timeout < 0:
                    raise TimeoutError("Operation timed out.")
                return await operation(*args, timeout=absolute_timeout, **kwargs)
            except AMQPException as exc:
                if not self._retry_policy.is_retryable(exc):
                    raise
                if absolute_timeout >= 0:
                    retry_active = self._retry_policy.increment(retry_settings, exc)
                    if not retry_active:
                        break
                    await asyncio.sleep(self._retry_policy.get_backoff_time(retry_settings, exc))
                    if exc.condition == ErrorCondition.LinkDetachForced:
                        await self._close_link_async()  # if link level error, close and open a new link
                    if exc.condition in (ErrorCondition.ConnectionCloseForced, ErrorCondition.SocketError):
                        # if connection detach or socket error, close and open a new connection
                        await self.close_async()
            finally:
                end_time = time.time()
                if absolute_timeout > 0:
                    absolute_timeout -= end_time - start_time
        raise retry_settings["history"][-1]

    async def open_async(self, connection=None):
        """Asynchronously open the client. The client can create a new Connection
        or an existing Connection can be passed in. This existing Connection
        may have an existing CBS authentication Session, which will be
        used for this client as well. Otherwise a new Session will be
        created.

        :param connection: An existing Connection that may be shared between
         multiple clients.
        :type connection: ~pyamqp.aio.Connection
        """
        # pylint: disable=protected-access
        if self._session:
            return  # already open.
        if connection:
            self._connection = connection
            self._external_connection = True
        if not self._connection:
            self._connection = Connection(
                "amqps://" + self._hostname if self._use_tls else "amqp://" + self._hostname,
                sasl_credential=self._auth.sasl,
                ssl_opts=self._ssl_opts,
                container_id=self._name,
                max_frame_size=self._max_frame_size,
                channel_max=self._channel_max,
                idle_timeout=self._idle_timeout,
                properties=self._properties,
                network_trace=self._network_trace,
                transport_type=self._transport_type,
                http_proxy=self._http_proxy,
                custom_endpoint_address=self._custom_endpoint_address,
                socket_timeout=self._socket_timeout,
                use_tls=self._use_tls,
            )
            await self._connection.open()
        if not self._session:
            self._session = self._connection.create_session(
                incoming_window=self._incoming_window, outgoing_window=self._outgoing_window
            )
            await self._session.begin()
        if self._auth.auth_type == AUTH_TYPE_CBS:
            self._cbs_authenticator = CBSAuthenticator(
                session=self._session, auth=self._auth, auth_timeout=self._auth_timeout
            )
            await self._cbs_authenticator.open()
        self._network_trace_params["amqpConnection"] = self._connection._container_id
        self._network_trace_params["amqpSession"] = self._session.name
        self._shutdown = False

        if self._keep_alive_interval:
            self._keep_alive_thread = asyncio.ensure_future(self._keep_alive_async())

    async def close_async(self):
        """Close the client asynchronously. This includes closing the Session
        and CBS authentication layer as well as the Connection.
        If the client was opened using an external Connection,
        this will be left intact.
        """
        self._shutdown = True
        if not self._session:
            return  # already closed.
        await self._close_link_async()
        if self._cbs_authenticator:
            await self._cbs_authenticator.close()
            self._cbs_authenticator = None
        await self._session.end()
        self._session = None
        if not self._external_connection:
            await self._connection.close()
            self._connection = None
        if self._keep_alive_thread:
            await self._keep_alive_thread
            self._keep_alive_thread = None
        self._network_trace_params["amqpConnection"] = ""
        self._network_trace_params["amqpSession"] = ""

    async def auth_complete_async(self):
        """Whether the authentication handshake is complete during
        connection initialization.

        :return: Whether the authentication handshake is complete.
        :rtype: bool
        """
        if self._cbs_authenticator and not await self._cbs_authenticator.handle_token():
            await self._connection.listen(wait=self._socket_timeout)
            return False
        return True

    async def client_ready_async(self):
        """
        Whether the handler has completed all start up processes such as
        establishing the connection, session, link and authentication, and
        is not ready to process messages.

        :return: Whether the client is ready to process messages.
        :rtype: bool
        """
        if not await self.auth_complete_async():
            return False
        if not await self._client_ready_async():
            try:
                await self._connection.listen(wait=self._socket_timeout)
            except ValueError:
                return True
            return False
        return True

    async def do_work_async(self, **kwargs):
        """Run a single connection iteration asynchronously.
        This will return `True` if the connection is still open
        and ready to be used for further work, or `False` if it needs
        to be shut down.

        :return: Whether the connection is still open and ready for work.
        :rtype: bool
        :raises: TimeoutError if CBS authentication timeout reached.
        """

        if self._shutdown:
            return False
        if not await self.client_ready_async():
            return True
        return await self._client_run_async(**kwargs)

    async def mgmt_request_async(
        self,
        message,
        *,
        operation: Optional[Union[bytes, str]] = None,
        operation_type: Optional[Union[bytes, str]] = None,
        node: str = "$management",
        timeout: float = 0,
        **kwargs
    ):
        """
        :param message: The message to send in the management request.
        :type message: ~pyamqp.message.Message
        :keyword str operation: The type of operation to be performed. This value will
         be service-specific, but common values include READ, CREATE and UPDATE.
         This value will be added as an application property on the message.
        :keyword str operation_type: The type on which to carry out the operation. This will
         be specific to the entities of the service. This value will be added as
         an application property on the message.
        :keyword str node: The target node. Default node is `$management`.
        :keyword float timeout: Provide an optional timeout in seconds within which a response
         to the management request must be received.
        :return: The response to the management request.
        :rtype: ~pyamqp.message.Message
        """

        # The method also takes "status_code_field" and "status_description_field"
        # keyword arguments as alternate names for the status code and description
        # in the response body. Those two keyword arguments are used in Azure services only.
        async with self._mgmt_link_lock_async:
            try:
                mgmt_link = self._mgmt_links[node]
            except KeyError:
                mgmt_link = ManagementOperation(self._session, endpoint=node, **kwargs)
                self._mgmt_links[node] = mgmt_link
                await mgmt_link.open()

        while not await self.client_ready_async():
            await asyncio.sleep(0.05)

        while not await mgmt_link.ready():
            await self._connection.listen(wait=False)

        operation_type = operation_type or b"empty"
        status, description, response = await mgmt_link.execute(
            message, operation=operation, operation_type=operation_type, timeout=timeout
        )
        return status, description, response


class SendClientAsync(SendClientSync, AMQPClientAsync):
    """An asynchronous AMQP client.

    :param target: The target AMQP service endpoint. This can either be the URI as
     a string or a ~pyamqp.endpoint.Target object.
    :type target: str, bytes or ~pyamqp.endpoint.Target
    :keyword auth: Authentication for the connection. This should be one of the following:
        - pyamqp.authentication.SASLAnonymous
        - pyamqp.authentication.SASLPlain
        - pyamqp.authentication.SASTokenAuth
        - pyamqp.authentication.JWTTokenAuth
     If no authentication is supplied, SASLAnnoymous will be used by default.
    :paramtype auth: ~pyamqp.authentication
    :keyword client_name: The name for the client, also known as the Container ID.
     If no name is provided, a random GUID will be used.
    :paramtype client_name: str or bytes
    :keyword network_trace: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :paramtype network_trace: bool
    :keyword retry_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :paramtype retry_policy: ~pyamqp.error.RetryPolicy
    :keyword keep_alive_interval: If set, a thread will be started to keep the connection
     alive during periods of user inactivity. The value will determine how long the
     thread will sleep (in seconds) between pinging the connection. If 0 or None, no
     thread will be started.
    :paramtype keep_alive_interval: int
    :keyword max_frame_size: Maximum AMQP frame size. Default is 63488 bytes.
    :paramtype max_frame_size: int
    :keyword channel_max: Maximum number of Session channels in the Connection.
    :paramtype channel_max: int
    :keyword idle_timeout: Timeout in seconds after which the Connection will close
     if there is no further activity.
    :paramtype idle_timeout: int
    :keyword auth_timeout: Timeout in seconds for CBS authentication. Otherwise this value will be ignored.
     Default value is 60s.
    :paramtype auth_timeout: int
    :keyword properties: Connection properties.
    :paramtype properties: dict[str, any]
    :keyword remote_idle_timeout_empty_frame_send_ratio: Ratio of empty frames to
     idle time for Connections with no activity. Value must be between
     0.0 and 1.0 inclusive. Default is 0.5.
    :paramtype remote_idle_timeout_empty_frame_send_ratio: float
    :keyword incoming_window: The size of the allowed window for incoming messages.
    :paramtype incoming_window: int
    :keyword outgoing_window: The size of the allowed window for outgoing messages.
    :paramtype outgoing_window: int
    :keyword handle_max: The maximum number of concurrent link handles.
    :paramtype handle_max: int
    :keyword on_attach: A callback function to be run on receipt of an ATTACH frame.
     The function must take 4 arguments: source, target, properties and error.
    :paramtype on_attach: func[
     ~pyamqp.endpoint.Source, ~pyamqp.endpoint.Target, dict, ~pyamqp.error.AMQPConnectionError]
    :keyword send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :paramtype send_settle_mode: ~pyamqp.constants.SenderSettleMode
    :keyword receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :paramtype receive_settle_mode: ~pyamqp.constants.ReceiverSettleMode
    :keyword desired_capabilities: The extension capabilities desired from the peer endpoint.
    :paramtype desired_capabilities: list[bytes]
    :keyword max_message_size: The maximum allowed message size negotiated for the Link.
    :paramtype max_message_size: int
    :keyword link_properties: Metadata to be sent in the Link ATTACH frame.
    :paramtype link_properties: dict[str, any]
    :keyword link_credit: The Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
     The default is 300.
    :paramtype link_credit: int
    :keyword transport_type: The type of transport protocol that will be used for communicating with
     the service. Default is `TransportType.Amqp` in which case port 5671 is used.
     If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
     be used instead which uses port 443 for communication.
    :paramtype transport_type: ~pyamqp.constants.TransportType
    :keyword http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
     Additionally the following keys may also be present: `'username', 'password'`.
    :paramtype http_proxy: dict[str, str]
    :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
     the Event Hubs service, allowing network requests to be routed through any application gateways or
     other paths needed for the host environment. Default is None.
     Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
     If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
    :paramtype custom_endpoint_address: str
    :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
     authenticate the identity of the connection endpoint. Ignored if ssl_context passed in. Default is None
     in which case `certifi.where()` will be used.
    :paramtype connection_verify: str
    :keyword ssl_context: An instance of ssl.SSLContext to be used. If this is specified, connection_verify is ignored.
    :paramtype ssl_context: ssl.SSLContext or None
    """

    async def _client_ready_async(self):
        """Determine whether the client is ready to start receiving messages.
        To be ready, the connection must be open and authentication complete,
        The Session, Link and MessageReceiver must be open and in non-errored
        states.

        :return: Whether or not the client is ready to start sending messages.
        :rtype: bool
        """
        if not self._link:
            self._link = self._session.create_sender_link(
                target_address=self.target,
                link_credit=self._link_credit,
                send_settle_mode=self._send_settle_mode,
                rcv_settle_mode=self._receive_settle_mode,
                max_message_size=self._max_message_size,
                properties=self._link_properties,
            )
            await self._link.attach()
            return False
        if self._link.get_state().value != 3:  # ATTACHED
            return False
        return True

    async def _client_run_async(self, **kwargs):
        """MessageSender Link is now open - perform message send
        on all pending messages.
        Will return True if operation successful and client can remain open for
        further work.

        :return: Whether or not the client should remain open for further work.
        :rtype: bool
        """
        await self._link.update_pending_deliveries()
        await self._connection.listen(wait=self._socket_timeout, **kwargs)
        return True

    async def _transfer_message_async(self, message_delivery, timeout=0):
        message_delivery.state = MessageDeliveryState.WaitingForSendAck
        on_send_complete = partial(self._on_send_complete_async, message_delivery)
        delivery = await self._link.send_transfer(
            message_delivery.message, on_send_complete=on_send_complete, timeout=timeout, send_async=True
        )
        return delivery

    async def _on_send_complete_async(self, message_delivery, reason, state):
        message_delivery.reason = reason
        if reason == LinkDeliverySettleReason.DISPOSITION_RECEIVED:
            if state and SEND_DISPOSITION_ACCEPT in state:
                message_delivery.state = MessageDeliveryState.Ok
            else:
                try:
                    error_info = state[SEND_DISPOSITION_REJECT]
                    self._process_send_error(
                        message_delivery,
                        condition=error_info[0][0],
                        description=error_info[0][1],
                        info=error_info[0][2],
                    )
                except TypeError:
                    self._process_send_error(message_delivery, condition=ErrorCondition.UnknownError)
        elif reason == LinkDeliverySettleReason.SETTLED:
            message_delivery.state = MessageDeliveryState.Ok
        elif reason == LinkDeliverySettleReason.TIMEOUT:
            message_delivery.state = MessageDeliveryState.Timeout
            message_delivery.error = TimeoutError("Sending message timed out.")
        else:
            # NotDelivered and other unknown errors
            self._process_send_error(message_delivery, condition=ErrorCondition.UnknownError)

    async def _send_message_impl_async(self, message, *, timeout: float = 0):
        expire_time = (time.time() + timeout) if timeout else None
        await self.open_async()
        message_delivery = _MessageDelivery(message, MessageDeliveryState.WaitingToBeSent, expire_time)

        while not await self.client_ready_async():
            await asyncio.sleep(0.05)

        await self._transfer_message_async(message_delivery, timeout)

        running = True
        while running and message_delivery.state not in MESSAGE_DELIVERY_DONE_STATES:
            running = await self.do_work_async()
        if message_delivery.state not in MESSAGE_DELIVERY_DONE_STATES:
            raise MessageException(
                condition=ErrorCondition.ClientError, description="Send failed - connection not running."
            )

        if message_delivery.state in (
            MessageDeliveryState.Error,
            MessageDeliveryState.Cancelled,
            MessageDeliveryState.Timeout,
        ):
            try:
                raise message_delivery.error  # pylint: disable=raising-bad-type
            except TypeError:
                # This is a default handler
                raise MessageException(condition=ErrorCondition.UnknownError, description="Send failed.") from None

    async def send_message_async(self, message, *, timeout: float = 0, **kwargs):
        """
        :param ~pyamqp.message.Message message: The message to send.
        :keyword float timeout: timeout in seconds. If set to
         0, the client will continue to wait until the message is sent or error happens. The
         default is 0.
        """
        await self._do_retryable_operation_async(
            self._send_message_impl_async, message=message, timeout=timeout, **kwargs
        )


class ReceiveClientAsync(ReceiveClientSync, AMQPClientAsync):
    """An asynchronous AMQP client.

    :param source: The source AMQP service endpoint. This can either be the URI as
     a string or a ~pyamqp.endpoint.Source object.
    :type source: str, bytes or ~pyamqp.endpoint.Source
    :keyword auth: Authentication for the connection. This should be one of the following:
        - pyamqp.authentication.SASLAnonymous
        - pyamqp.authentication.SASLPlain
        - pyamqp.authentication.SASTokenAuth
        - pyamqp.authentication.JWTTokenAuth
     If no authentication is supplied, SASLAnnoymous will be used by default.
    :paramtype auth: ~pyamqp.authentication
    :keyword client_name: The name for the client, also known as the Container ID.
     If no name is provided, a random GUID will be used.
    :paramtype client_name: str or bytes
    :keyword network_trace: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :paramtype network_trace: bool
    :keyword retry_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :paramtype retry_policy: ~pyamqp.error.RetryPolicy
    :keyword keep_alive_interval: If set, a thread will be started to keep the connection
     alive during periods of user inactivity. The value will determine how long the
     thread will sleep (in seconds) between pinging the connection. If 0 or None, no
     thread will be started.
    :paramtype keep_alive_interval: int
    :keyword max_frame_size: Maximum AMQP frame size. Default is 63488 bytes.
    :paramtype max_frame_size: int
    :keyword channel_max: Maximum number of Session channels in the Connection.
    :paramtype channel_max: int
    :keyword idle_timeout: Timeout in seconds after which the Connection will close
     if there is no further activity.
    :paramtype idle_timeout: int
    :keyword auth_timeout: Timeout in seconds for CBS authentication. Otherwise this value will be ignored.
     Default value is 60s.
    :paramtype auth_timeout: int
    :keyword properties: Connection properties.
    :paramtype properties: dict[str, any]
    :keyword remote_idle_timeout_empty_frame_send_ratio: Ratio of empty frames to
     idle time for Connections with no activity. Value must be between
     0.0 and 1.0 inclusive. Default is 0.5.
    :paramtype remote_idle_timeout_empty_frame_send_ratio: float
    :keyword incoming_window: The size of the allowed window for incoming messages.
    :paramtype incoming_window: int
    :keyword outgoing_window: The size of the allowed window for outgoing messages.
    :paramtype outgoing_window: int
    :keyword handle_max: The maximum number of concurrent link handles.
    :paramtype handle_max: int
    :keyword on_attach: A callback function to be run on receipt of an ATTACH frame.
     The function must take 4 arguments: source, target, properties and error.
    :paramtype on_attach: func[
     ~pyamqp.endpoint.Source, ~pyamqp.endpoint.Target, dict, ~pyamqp.error.AMQPConnectionError]
    :keyword send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :paramtype send_settle_mode: ~pyamqp.constants.SenderSettleMode
    :keyword receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :paramtype receive_settle_mode: ~pyamqp.constants.ReceiverSettleMode
    :keyword desired_capabilities: The extension capabilities desired from the peer endpoint.
    :paramtype desired_capabilities: list[bytes]
    :keyword max_message_size: The maximum allowed message size negotiated for the Link.
    :paramtype max_message_size: int
    :keyword link_properties: Metadata to be sent in the Link ATTACH frame.
    :paramtype link_properties: dict[str, any]
    :keyword link_credit: The Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
     The default is 300.
    :paramtype link_credit: int
    :keyword transport_type: The type of transport protocol that will be used for communicating with
     the service. Default is `TransportType.Amqp` in which case port 5671 is used.
     If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
     be used instead which uses port 443 for communication.
    :paramtype transport_type: ~pyamqp.constants.TransportType
    :keyword http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
     Additionally the following keys may also be present: `'username', 'password'`.
    :paramtype http_proxy: dict[str, str]
    :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
     the Event Hubs service, allowing network requests to be routed through any application gateways or
     other paths needed for the host environment. Default is None.
     Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
     If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
    :paramtype custom_endpoint_address: str
    :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
     authenticate the identity of the connection endpoint. Ignored if ssl_context passed in. Default is None
     in which case `certifi.where()` will be used.
    :paramtype connection_verify: str
    :keyword ssl_context: An instance of ssl.SSLContext to be used. If this is specified, connection_verify is ignored.
    :paramtype ssl_context: ssl.SSLContext or None
    """

    async def _client_ready_async(self):
        """Determine whether the client is ready to start receiving messages.
        To be ready, the connection must be open and authentication complete,
        The Session, Link and MessageReceiver must be open and in non-errored
        states.

        :return: Whether the client is ready to start receiving messages.
        :rtype: bool
        """
        if not self._link:
            self._link = self._session.create_receiver_link(
                source_address=self.source,
                link_credit=0,  # link_credit=0 on flow frame sent before client is ready
                send_settle_mode=self._send_settle_mode,
                rcv_settle_mode=self._receive_settle_mode,
                max_message_size=self._max_message_size,
                on_transfer=self._message_received_async,
                properties=self._link_properties,
                desired_capabilities=self._desired_capabilities,
                on_attach=self._on_attach,
            )
            await self._link.attach()
            return False
        if self._link.get_state().value != 3:  # ATTACHED
            return False
        return True

    async def _client_run_async(self, **kwargs):
        """MessageReceiver Link is now open - start receiving messages.
        Will return True if operation successful and client can remain open for
        further work.

        :return: Whether the client can remain open for further work.
        :rtype: bool
        """
        try:
            if self._link.current_link_credit<= 0:
                await self._link.flow(link_credit=self._link_credit)
            await self._connection.listen(wait=self._socket_timeout, **kwargs)
        except ValueError:
            _logger.info("Timeout reached, closing receiver.", extra=self._network_trace_params)
            self._shutdown = True
            return False
        return True

    async def _message_received_async(self, frame, message):
        """Callback run on receipt of every message. If there is
        a user-defined callback, this will be called.
        Additionally if the client is retrieving messages for a batch
        or iterator, the message will be added to an internal queue.

        :param tuple frame: Received frame.
        :param message: Received message.
        :type message: ~pyamqp.message.Message
        """
        self._last_activity_timestamp = time.time()
        if self._message_received_callback:
            await self._message_received_callback(message)
        if not self._streaming_receive:
            self._received_messages.put((frame, message))

    async def _receive_message_batch_impl_async(
        self,
        max_batch_size: Optional[int] = None,
        on_message_received: Optional[Callable] = None,
        timeout: float = 0,
    ):
        self._message_received_callback = on_message_received
        max_batch_size = max_batch_size or self._link_credit
        timeout_time = time.time() + timeout if timeout else 0
        receiving = True
        batch: List[Message] = []
        await self.open_async()
        while len(batch) < max_batch_size:
            try:
                # TODO: This drops the transfer frame data
                _, message = self._received_messages.get_nowait()
                batch.append(message)
                self._received_messages.task_done()
            except queue.Empty:
                break
        else:
            return batch

        to_receive_size = max_batch_size - len(batch)
        before_queue_size = self._received_messages.qsize()

        while receiving and to_receive_size > 0:
            now_time = time.time()
            if timeout_time and now_time > timeout_time:
                break

            try:
                receiving = await asyncio.wait_for(
                    self.do_work_async(batch=to_receive_size), timeout=timeout_time - now_time if timeout else None
                )
            except asyncio.TimeoutError:
                break

            cur_queue_size = self._received_messages.qsize()
            # after do_work, check how many new messages have been received since previous iteration
            received = cur_queue_size - before_queue_size
            if to_receive_size < max_batch_size and received == 0:
                # there are already messages in the batch, and no message is received in the current cycle
                # return what we have
                break

            to_receive_size -= received
            before_queue_size = cur_queue_size

        while len(batch) < max_batch_size:
            try:
                _, message = self._received_messages.get_nowait()
                batch.append(message)
                self._received_messages.task_done()
            except queue.Empty:
                break
        return batch

    async def close_async(self):
        self._received_messages = queue.Queue()
        await super(ReceiveClientAsync, self).close_async()

    async def receive_message_batch_async(  # pylint: disable=unused-argument
        self,
        *,
        max_batch_size: Optional[int] = None,
        on_message_received: Optional[Callable] = None,
        timeout: float = 0,
        **kwargs
    ) -> Coroutine[Any, Any, list]:
        """Receive a batch of messages. Messages returned in the batch have already been
        accepted - if you wish to add logic to accept or reject messages based on custom
        criteria, pass in a callback. This method will return as soon as some messages are
        available rather than waiting to achieve a specific batch size, and therefore the
        number of messages returned per call will vary up to the maximum allowed.

        :keyword max_batch_size: The maximum number of messages that can be returned in
         one call. This value cannot be larger than the prefetch value, and if not specified,
         the prefetch value will be used.
        :paramtype max_batch_size: int
        :keyword on_message_received: A callback to process messages as they arrive from the
         service. It takes a single argument, a ~pyamqp.message.Message object.
        :paramtype on_message_received: callable[~pyamqp.message.Message]
        :keyword timeout: Timeout in seconds for which to wait to receive any messages.
         If no messages are received in this time, an empty list will be returned. If set to
         0, the client will continue to wait until at least one message is received. The
         default is 0.
        :paramtype timeout: float
        :returns: Retryable operation coroutine.
        :rtype: Coroutine[Any, Any, list]
        """
        return await self._do_retryable_operation_async(
            self._receive_message_batch_impl_async,
            max_batch_size=max_batch_size,
            on_message_received=on_message_received,
            timeout=timeout,
        )

    async def receive_messages_iter_async(self, timeout=None, on_message_received=None):
        """Receive messages by generator. Messages returned in the generator have already been
        accepted - if you wish to add logic to accept or reject messages based on custom
        criteria, pass in a callback.

        :param on_message_received: A callback to process messages as they arrive from the
         service. It takes a single argument, a ~pyamqp.message.Message object.
        :type on_message_received: callable[~pyamqp.message.Message]
        :param float timeout: Timeout in seconds for which to wait to receive any messages.
        :return: A generator of messages.
        :rtype: generator[~pyamqp.message.Message]
        """
        self._message_received_callback = on_message_received
        return self._message_generator_async(timeout=timeout)

    async def _message_generator_async(self, timeout=None):
        """Iterate over processed messages in the receive queue.

        :param float timeout: Timeout in seconds for which to wait to receive any messages.
        :return: A generator of messages.
        :rtype: generator[~pyamqp.message.Message]
        """
        self.open()
        receiving = True
        message = None
        self._last_activity_timestamp = time.time()
        self._timeout_reached = False
        self._timeout = timeout if timeout else self._timeout
        try:
            while receiving and not self._timeout_reached:
                if self._timeout > 0:
                    if time.time() - self._last_activity_timestamp >= self._timeout:
                        self._timeout_reached = True

                if not self._timeout_reached:
                    receiving = await self.do_work_async()

                while not self._received_messages.empty():
                    message = self._received_messages.get()
                    self._last_activity_timestamp = time.time()
                    self._received_messages.task_done()
                    yield message

        finally:
            if self._shutdown:
                await self.close_async()

    @overload
    async def settle_messages_async(
        self,
        delivery_id: Union[int, Tuple[int, int]],
        delivery_tag: bytes,
        outcome: Literal["accepted"],
        *,
        batchable: Optional[bool] = None
    ): ...

    @overload
    async def settle_messages_async(
        self,
        delivery_id: Union[int, Tuple[int, int]],
        delivery_tag: bytes,
        outcome: Literal["released"],
        *,
        batchable: Optional[bool] = None
    ): ...

    @overload
    async def settle_messages_async(
        self,
        delivery_id: Union[int, Tuple[int, int]],
        delivery_tag: bytes,
        outcome: Literal["rejected"],
        *,
        error: Optional[AMQPError] = None,
        batchable: Optional[bool] = None
    ): ...

    @overload
    async def settle_messages_async(
        self,
        delivery_id: Union[int, Tuple[int, int]],
        delivery_tag: bytes,
        outcome: Literal["modified"],
        *,
        delivery_failed: Optional[bool] = None,
        undeliverable_here: Optional[bool] = None,
        message_annotations: Optional[Dict[Union[str, bytes], Any]] = None,
        batchable: Optional[bool] = None
    ): ...

    @overload
    async def settle_messages_async(
        self,
        delivery_id: Union[int, Tuple[int, int]],
        delivery_tag: bytes,
        outcome: Literal["received"],
        *,
        section_number: int,
        section_offset: int,
        batchable: Optional[bool] = None
    ): ...

    async def settle_messages_async(
        self, delivery_id: Union[int, Tuple[int, int]], delivery_tag: bytes, outcome: str, **kwargs
    ):
        batchable = kwargs.pop("batchable", None)
        if outcome.lower() == "accepted":
            state: Outcomes = Accepted()
        elif outcome.lower() == "released":
            state = Released()
        elif outcome.lower() == "rejected":
            state = Rejected(**kwargs)
        elif outcome.lower() == "modified":
            state = Modified(**kwargs)
        elif outcome.lower() == "received":
            state = Received(**kwargs)
        else:
            raise ValueError("Unrecognized message output: {}".format(outcome))
        try:
            first, last = cast(Tuple, delivery_id)
        except TypeError:
            first = delivery_id
            last = None
        await self._link.send_disposition(
            first_delivery_id=first,
            last_delivery_id=last,
            delivery_tag=delivery_tag,
            settled=True,
            delivery_state=state,
            batchable=batchable,
            wait=True,
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._test_base import _SendTest


class SendEventsTest(_SendTest):
    def __init__(self, arguments):
        super().__init__(arguments)

    def run_batch_sync(self):
        if self.args.batch_size > 1:
            self.producer.send_batch([self._build_event() for _ in range(self.args.batch_size)])
        else:
            self.producer.send_event(self._build_event())
        return self.args.batch_size

    async def run_batch_async(self):
        if self.args.batch_size > 1:
            await self.async_producer.send_batch([self._build_event() for _ in range(self.args.batch_size)])
        else:
            await self.async_producer.send_event(self._build_event())
        return self.args.batch_size



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import warnings
from typing import Optional, Dict, Any, Union, TYPE_CHECKING
from urllib.parse import urlparse

from azure.core.pipeline.policies import RetryMode
from ._constants import TransportType, DEFAULT_AMQPS_PORT, DEFAULT_AMQP_WSS_PORT

if TYPE_CHECKING:
    from ssl import SSLContext
    from ._transport._base import AmqpTransport
    from .aio._transport._base_async import AmqpTransportAsync


class Configuration:  # pylint:disable=too-many-instance-attributes
    def __init__(
        self,  # pylint:disable=unused-argument
        *,
        hostname: str,
        amqp_transport: Union["AmqpTransport", "AmqpTransportAsync"],
        socket_timeout: Optional[float] = None,
        user_agent: Optional[str] = None,
        retry_total: int = 3,
        retry_mode: Union[str, RetryMode] = RetryMode.Exponential,
        retry_backoff_factor: float = 0.8,
        retry_backoff_max: int = 120,
        network_tracing: bool = False,
        http_proxy: Optional[Dict[str, Any]] = None,
        transport_type: Optional[TransportType] = None,
        auth_timeout: int = 60,
        prefetch: int = 300,
        max_batch_size: int = 300,
        receive_timeout: int = 0,
        send_timeout: int = 60,
        custom_endpoint_address: Optional[str] = None,
        connection_verify: Optional[str] = None,
        ssl_context: Optional["SSLContext"] = None,
        use_tls: bool = True,
        **kwargs: Any
    ):
        self.user_agent = user_agent
        self.retry_total = retry_total
        self.max_retries = self.retry_total
        self.retry_mode = retry_mode
        self.backoff_factor = retry_backoff_factor
        self.backoff_max = retry_backoff_max
        self.network_tracing = network_tracing
        self.http_proxy = http_proxy
        self.transport_type = TransportType.AmqpOverWebsocket if self.http_proxy else transport_type
        # if transport_type is not provided, it is None, we will default to Amqp
        self.transport_type = self.transport_type or TransportType.Amqp
        self.auth_timeout = auth_timeout
        self.prefetch = prefetch
        self.max_batch_size = max_batch_size
        self.receive_timeout = receive_timeout
        self.send_timeout = send_timeout
        self.custom_endpoint_address = custom_endpoint_address
        self.connection_verify = connection_verify
        self.ssl_context = ssl_context
        if self.ssl_context and self.connection_verify:
            warnings.warn("ssl_context is specified, connection_verify will be ignored.")
        self.custom_endpoint_hostname = None
        self.hostname = hostname
        self.use_tls = use_tls

        if self.http_proxy or self.transport_type.value == TransportType.AmqpOverWebsocket.value:
            self.transport_type = TransportType.AmqpOverWebsocket
            self.connection_port = DEFAULT_AMQP_WSS_PORT
            self.socket_timeout = socket_timeout or 1
            if amqp_transport.KIND == "pyamqp":
                self.hostname += "/$servicebus/websocket"
        else:
            self.socket_timeout = socket_timeout or 0.2
            self.connection_port = DEFAULT_AMQPS_PORT

        # custom end point
        if self.custom_endpoint_address:
            # if the custom_endpoint_address doesn't include the schema,
            # we prepend a default one to make urlparse work
            if self.custom_endpoint_address.find("//") == -1:
                self.custom_endpoint_address = "sb://" + self.custom_endpoint_address
            endpoint = urlparse(self.custom_endpoint_address)
            # this line is to maintain backward compatibility for custom endpoint
            # if the transport_type is not provided, we will default to AmqpOverWebsocket
            self.transport_type = transport_type or TransportType.AmqpOverWebsocket
            self.custom_endpoint_hostname = endpoint.hostname
            if amqp_transport.KIND == "pyamqp":
                # if transport is amqp, we will not append this path to the final endpoint
                self.custom_endpoint_address += "/$servicebus/websocket"
            # in case proxy and custom endpoint are both provided, we default port to 443 if it's not provided
            self.connection_port = endpoint.port or DEFAULT_AMQP_WSS_PORT



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub for a period of time asynchronously.
"""

import asyncio
import os
import time
from azure.eventhub.aio import EventHubConsumerClient
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]
RECEIVE_DURATION = 15


async def on_event(partition_context, event):
    # Put your code here.
    print("Received event from partition: {}.".format(partition_context.partition_id))
    await partition_context.update_checkpoint(event)


async def on_partition_initialize(partition_context):
    # Put your code here.
    print("Partition: {} has been initialized.".format(partition_context.partition_id))


async def on_partition_close(partition_context, reason):
    # Put your code here.
    print("Partition: {} has been closed, reason for closing: {}.".format(partition_context.partition_id, reason))


async def on_error(partition_context, error):
    # Put your code here. partition_context can be None in the on_error callback.
    if partition_context:
        print(
            "An exception: {} occurred during receiving from Partition: {}.".format(
                partition_context.partition_id, error
            )
        )
    else:
        print("An exception: {} occurred during the load balance process.".format(error))


async def main():
    client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        consumer_group="$default",
    )

    print("Consumer will keep receiving for {} seconds, start time is {}.".format(RECEIVE_DURATION, time.time()))

    async with client:
        task = asyncio.ensure_future(
            client.receive(
                on_event=on_event,
                on_error=on_error,
                on_partition_close=on_partition_close,
                on_partition_initialize=on_partition_initialize,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        )
        await asyncio.sleep(RECEIVE_DURATION)
    await task

    print("Consumer has stopped receiving, end time is {}.".format(time.time()))


if __name__ == "__main__":
    asyncio.run(main())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# The Messaging layer defines a concrete set of delivery states which can be used (via the disposition frame)
# to indicate the state of the message at the receiver.

# Delivery states may be either terminal or non-terminal. Once a delivery reaches a terminal delivery-state,
# the state for that delivery will no longer change. A terminal delivery-state is referred to as an outcome.

# The following outcomes are formally defined by the messaging layer to indicate the result of processing at the
# receiver:

#     - accepted: indicates successful processing at the receiver
#     - rejected: indicates an invalid and unprocessable message
#     - released: indicates that the message was not (and will not be) processed
#     - modified: indicates that the message was modified, but not processed

# The following non-terminal delivery-state is formally defined by the messaging layer for use during link
# recovery to allow the sender to resume the transfer of a large message without retransmitting all the
# message data:

#     - received: indicates partial message data seen by the receiver as well as the starting point for a
#         resumed transfer

# TODO: fix mypy errors for _code/_definition/__defaults__ (issue #26500)
from collections import namedtuple

from .types import AMQPTypes, FieldDefinition, ObjDefinition
from .constants import FIELD
from .performatives import _CAN_ADD_DOCSTRING


Received = namedtuple("Received", ["section_number", "section_offset"])
Received._code = 0x00000023  # type: ignore # pylint:disable=protected-access
Received._definition = (  # type: ignore # pylint:disable=protected-access
    FIELD("section_number", AMQPTypes.uint, True, None, False),
    FIELD("section_offset", AMQPTypes.ulong, True, None, False),
)
if _CAN_ADD_DOCSTRING:
    Received.__doc__ = """
    At the target the received state indicates the furthest point in the payload of the message
    which the target will not need to have resent if the link is resumed. At the source the received state represents
    the earliest point in the payload which the Sender is able to resume transferring at in the case of link
    resumption. When resuming a delivery, if this state is set on the first transfer performative it indicates
    the offset in the payload at which the first resumed delivery is starting. The Sender MUST NOT send the
    received state on transfer or disposition performatives except on the first transfer performative on a
    resumed delivery.

    :param int section_number:
        When sent by the Sender this indicates the first section of the message (with sectionnumber 0 being the
        first section) for which data can be resent. Data from sections prior to the given section cannot be
        retransmitted for this delivery. When sent by the Receiver this indicates the first section of the message
        for which all data may not yet have been received.
    :param int section_offset:
        When sent by the Sender this indicates the first byte of the encoded section data of the section given by
        section-number for which data can be resent (with section-offset 0 being the first byte). Bytes from the
        same section prior to the given offset section cannot be retransmitted for this delivery. When sent by the
        Receiver this indicates the first byte of the given section which has not yet been received. Note that if
        a receiver has received all of section number X (which contains N bytes of data), but none of section
        number X + 1, then it may indicate this by sending either Received(section-number=X, section-offset=N) or
        Received(section-number=X+1, section-offset=0). The state Received(sectionnumber=0, section-offset=0)
        indicates that no message data at all has been transferred.
    """


Accepted = namedtuple("Accepted", [])
Accepted._code = 0x00000024  # type: ignore # pylint:disable=protected-access
Accepted._definition = ()  # type: ignore # pylint:disable=protected-access
if _CAN_ADD_DOCSTRING:
    Accepted.__doc__ = """
    The accepted outcome.

    At the source the accepted state means that the message has been retired from the node, and transfer of
    payload data will not be able to be resumed if the link becomes suspended. A delivery may become accepted at
    the source even before all transfer frames have been sent, this does not imply that the remaining transfers
    for the delivery will not be sent - only the aborted fiag on the transfer performative can be used to indicate
    a premature termination of the transfer. At the target, the accepted outcome is used to indicate that an
    incoming Message has been successfully processed, and that the receiver of the Message is expecting the sender
    to transition the delivery to the accepted state at the source. The accepted outcome does not increment the
    delivery-count in the header of the accepted Message.
    """


Rejected = namedtuple("Rejected", ["error"])
Rejected.__new__.__defaults__ = (None,) * len(Rejected._fields)  # type: ignore
Rejected._code = 0x00000025  # type: ignore # pylint:disable=protected-access
Rejected._definition = (FIELD("error", ObjDefinition.error, False, None, False),)  # type: ignore # pylint:disable=protected-access
if _CAN_ADD_DOCSTRING:
    Rejected.__doc__ = """
    The rejected outcome.

    At the target, the rejected outcome is used to indicate that an incoming Message is invalid and therefore
    unprocessable. The rejected outcome when applied to a Message will cause the delivery-count to be incremented
    in the header of the rejected Message. At the source, the rejected outcome means that the target has informed
    the source that the message was rejected, and the source has taken the required action. The delivery SHOULD
    NOT ever spontaneously attain the rejected state at the source.

    :param ~uamqp.error.AMQPError error: The error that caused the message to be rejected.
        The value supplied in this field will be placed in the delivery-annotations of the rejected Message
        associated with the symbolic key "rejected".
    """


Released = namedtuple("Released", [])
Released._code = 0x00000026  # type: ignore # pylint:disable=protected-access
Released._definition = ()  # type: ignore # pylint:disable=protected-access
if _CAN_ADD_DOCSTRING:
    Released.__doc__ = """
    The released outcome.

    At the source the released outcome means that the message is no longer acquired by the receiver, and has been
    made available for (re-)delivery to the same or other targets receiving from the node. The message is unchanged
    at the node (i.e. the delivery-count of the header of the released Message MUST NOT be incremented).
    As released is a terminal outcome, transfer of payload data will not be able to be resumed if the link becomes
    suspended. A delivery may become released at the source even before all transfer frames have been sent, this
    does not imply that the remaining transfers for the delivery will not be sent. The source MAY spontaneously
    attain the released outcome for a Message (for example the source may implement some sort of time bound
    acquisition lock, after which the acquisition of a message at a node is revoked to allow for delivery to an
    alternative consumer).

    At the target, the released outcome is used to indicate that a given transfer was not and will not be acted upon.
    """


Modified = namedtuple("Modified", ["delivery_failed", "undeliverable_here", "message_annotations"])
Modified.__new__.__defaults__ = (None,) * len(Modified._fields)  # type: ignore
Modified._code = 0x00000027  # type: ignore # pylint:disable=protected-access
Modified._definition = (  # type: ignore # pylint:disable=protected-access
    FIELD("delivery_failed", AMQPTypes.boolean, False, None, False),
    FIELD("undeliverable_here", AMQPTypes.boolean, False, None, False),
    FIELD("message_annotations", FieldDefinition.fields, False, None, False),
)
if _CAN_ADD_DOCSTRING:
    Modified.__doc__ = """
    The modified outcome.

    At the source the modified outcome means that the message is no longer acquired by the receiver, and has been
    made available for (re-)delivery to the same or other targets receiving from the node. The message has been
    changed at the node in the ways indicated by the fields of the outcome. As modified is a terminal outcome,
    transfer of payload data will not be able to be resumed if the link becomes suspended. A delivery may become
    modified at the source even before all transfer frames have been sent, this does not imply that the remaining
    transfers for the delivery will not be sent. The source MAY spontaneously attain the modified outcome for a
    Message (for example the source may implement some sort of time bound acquisition lock, after which the
    acquisition of a message at a node is revoked to allow for delivery to an alternative consumer with the
    message modified in some way to denote the previous failed, e.g. with delivery-failed set to true).
    At the target, the modified outcome is used to indicate that a given transfer was not and will not be acted
    upon, and that the message should be modified in the specified ways at the node.

    :param bool delivery_failed: Count the transfer as an unsuccessful delivery attempt.
        If the delivery-failed fiag is set, any Messages modified MUST have their deliverycount incremented.
    :param bool undeliverable_here: Prevent redelivery.
        If the undeliverable-here is set, then any Messages released MUST NOT be redelivered to the modifying
        Link Endpoint.
    :param dict message_annotations: Message attributes.
        Map containing attributes to combine with the existing message-annotations held in the Message's header
        section. Where the existing message-annotations of the Message contain an entry with the same key as an
        entry in this field, the value in this field associated with that key replaces the one in the existing
        headers; where the existing message-annotations has no such value, the value in this map is added.
    """
